# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import json
import pathlib
import re
import time
import typing
from collections import OrderedDict
from copy import copy, deepcopy
from datetime import datetime
from os import environ
from typing import Any, Optional, Union

import pydantic.v1.error_wrappers

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas.notification
import mlrun.common.secrets
import mlrun.utils.regex

from .utils import (
    dict_to_json,
    dict_to_yaml,
    get_artifact_target,
    logger,
    template_artifact_path,
)

# Changing {run_id} will break and will not be backward compatible.
RUN_ID_PLACE_HOLDER = "{run_id}"  # IMPORTANT: shouldn't be changed.


class ModelObj:
    _dict_fields = []
    # Bellow attributes are used in to_dict method
    # Fields to strip from the object by default if strip=True
    _default_fields_to_strip = []
    # Fields that will be serialized by the object's _serialize_field method
    _fields_to_serialize = []
    # Fields that will be enriched by the object's _enrich_field method
    _fields_to_enrich = []
    # Fields that will be ignored by the object's _is_valid_field_value_for_serialization method
    _fields_to_skip_validation = []

    @staticmethod
    def _verify_list(param, name):
        if not isinstance(param, list):
            raise ValueError(f"Parameter {name} must be a list")

    @staticmethod
    def _verify_dict(param, name, new_type=None):
        if (
            param is not None
            and not isinstance(param, dict)
            and not hasattr(param, "to_dict")
        ):
            raise ValueError(f"Parameter {name} must be a dict or object")
        if new_type and (isinstance(param, dict) or param is None):
            return new_type.from_dict(param)
        return param

    @mlrun.utils.filter_warnings("ignore", FutureWarning)
    def to_dict(
        self,
        fields: Optional[list] = None,
        exclude: Optional[list] = None,
        strip: bool = False,
    ) -> dict:
        """
        Convert the object to a dict

        :param fields:  A list of fields to include in the dictionary. If not provided, the default value is taken
            from `self._dict_fields` or from the object __init__ params.
        :param exclude: A list of fields to exclude from the dictionary.
        :param strip:  If True, the object's `_default_fields_to_strip` attribute is appended to the exclude list.
            Strip purpose is to remove fields that are context / environment specific and not required for actually
            define the object.

        :return: A dictionary representation of the object.
        """
        struct = {}

        fields = self._resolve_initial_to_dict_fields(fields)
        fields_to_exclude = exclude or []
        if strip:
            fields_to_exclude += self._default_fields_to_strip

        # fields_to_save is built from the fields list minus the fields to exclude minus the fields that requires
        # serialization and enrichment (because they will be added later to the struct)
        fields_to_save = (
            set(fields)
            - set(fields_to_exclude)
            - set(self._fields_to_serialize)
            - set(self._fields_to_enrich)
        )

        # Iterating over the fields to save and adding them to the struct
        for field_name in fields_to_save:
            field_value = getattr(self, field_name, None)
            if self._is_valid_field_value_for_serialization(
                field_name, field_value, strip
            ):
                # If the field value has attribute to_dict, we call it.
                # If one of the attributes is a third party object that has to_dict method (such as k8s objects), then
                # add it to the object's _fields_to_serialize attribute and handle it in the _serialize_field method.
                if hasattr(field_value, "to_dict"):
                    # TODO: Allow passing fields to exclude from the parent object to the child object
                    #  e.g.: run.to_dict(exclude=["status.artifacts"])
                    field_value = field_value.to_dict(strip=strip)
                    if self._is_valid_field_value_for_serialization(
                        field_name, field_value, strip
                    ):
                        struct[field_name] = field_value
                else:
                    struct[field_name] = field_value

        # Subtracting the fields_to_exclude from the fields_to_serialize because if we want to exclude a field there
        # is no need to serialize it.
        fields_to_serialize = list(
            set(self._fields_to_serialize) - set(fields_to_exclude)
        )
        self._resolve_field_value_by_method(
            struct, self._serialize_field, fields_to_serialize, strip
        )

        # Subtracting the fields_to_exclude from the fields_to_enrich because if we want to exclude a field there
        # is no need to enrich it.
        fields_to_enrich = list(set(self._fields_to_enrich) - set(fields_to_exclude))
        self._resolve_field_value_by_method(
            struct, self._enrich_field, fields_to_enrich, strip
        )

        self._apply_enrichment_before_to_dict_completion(struct, strip=strip)
        return struct

    def _resolve_initial_to_dict_fields(self, fields: Optional[list] = None) -> list:
        """
        Resolve fields to be used in to_dict method.
        If fields is None, use `_dict_fields` attribute of the object.
        If fields is None and `_dict_fields` is empty, use the object's __init__ parameters.
        :param fields: List of fields to iterate over.

        :return: List of fields to iterate over.
        """
        return (
            fields
            or self._dict_fields
            or list(inspect.signature(self.__init__).parameters.keys())
        )

    def _is_valid_field_value_for_serialization(
        self, field_name: str, field_value: str, strip: bool = False
    ) -> bool:
        """
        Check if the field value is valid for serialization.
        If field name is in `_fields_to_skip_validation` attribute, skip validation and return True.
        If strip is False skip validation and return True.
        If field value is None or empty dict/list, then no need to store it.
        :param field_name:  Field name.
        :param field_value: Field value.

        :return: True if the field value is valid for serialization, False otherwise.
        """
        if field_name in self._fields_to_skip_validation:
            return True
        # TODO: remove when Runtime initialization will be refactored and enrichment will be moved to BE
        # if not strip:
        #     return True

        return field_value is not None and not (
            (isinstance(field_value, dict) or isinstance(field_value, list))
            and not field_value
        )

    def _resolve_field_value_by_method(
        self,
        struct: dict,
        method: typing.Callable,
        fields: Optional[typing.Union[list, set]] = None,
        strip: bool = False,
    ) -> dict:
        for field_name in fields:
            field_value = method(struct=struct, field_name=field_name, strip=strip)
            if self._is_valid_field_value_for_serialization(
                field_name, field_value, strip
            ):
                struct[field_name] = field_value
        return struct

    def _serialize_field(
        self, struct: dict, field_name: Optional[str] = None, strip: bool = False
    ) -> typing.Any:
        # We pull the field from self and not from struct because it was excluded from the struct when looping over
        # the fields to save.
        return getattr(self, field_name, None)

    def _enrich_field(
        self, struct: dict, field_name: Optional[str] = None, strip: bool = False
    ) -> typing.Any:
        # We first try to pull from struct because the field might have been already serialized and if not,
        # we pull from self
        return struct.get(field_name, None) or getattr(self, field_name, None)

    def _apply_enrichment_before_to_dict_completion(
        self, struct: dict, strip: bool = False
    ) -> dict:
        return struct

    @classmethod
    def from_dict(
        cls,
        struct=None,
        fields=None,
        deprecated_fields: Optional[dict] = None,
        init_with_params: bool = False,
    ):
        """create an object from a python dictionary"""
        struct = {} if struct is None else copy(struct)
        deprecated_fields = deprecated_fields or {}
        fields = fields or cls._dict_fields
        if not fields:
            fields = list(inspect.signature(cls.__init__).parameters.keys())

        if init_with_params:
            kwargs = {field: struct.pop(field, None) for field in fields}
            kwargs.pop("self", None)
            new_obj = cls(**kwargs)
        else:
            new_obj = cls()

        if struct:
            # we are looping over the fields to save the same order and behavior in which the class
            # initialize the attributes
            for field in fields:
                # we want to set the field only if the field exists in struct
                if field in struct:
                    field_val = struct.get(field, None)
                    if field not in deprecated_fields:
                        setattr(new_obj, field, field_val)

            for deprecated_field, new_field in deprecated_fields.items():
                field_value = struct.get(new_field) or struct.get(deprecated_field)
                if field_value:
                    setattr(new_obj, new_field, field_value)

        return new_obj

    def to_yaml(self, exclude=None, strip: bool = False) -> str:
        """convert the object to yaml

        :param exclude: list of fields to exclude from the yaml
        :param strip:   if True, strip fields that are not required for actually define the object
        """
        return dict_to_yaml(self.to_dict(exclude=exclude, strip=strip))

    def to_json(self, exclude=None, strip: bool = False):
        """convert the object to json

        :param exclude: list of fields to exclude from the json
        :param strip:   if True, strip fields that are not required for actually define the object
        """
        return dict_to_json(self.to_dict(exclude=exclude, strip=strip))

    def to_str(self):
        """convert the object to string (with dict layout)"""
        return self.__str__()

    def __str__(self):
        return str(self.to_dict())

    def copy(self):
        """create a copy of the object"""
        return deepcopy(self)


# model class for building ModelObj dictionaries
class ObjectDict:
    kind = "object_dict"

    def __init__(self, classes_map, default_kind=""):
        self._children = OrderedDict()
        self._default_kind = default_kind
        self._classes_map = classes_map

    def values(self):
        return self._children.values()

    def keys(self):
        return self._children.keys()

    def items(self):
        return self._children.items()

    def __len__(self):
        return len(self._children)

    def __iter__(self):
        yield from self._children.keys()

    def __getitem__(self, name):
        return self._children[name]

    def __setitem__(self, key, item):
        self._children[key] = self._get_child_object(item, key)

    def __delitem__(self, key):
        del self._children[key]

    def update(self, key, item):
        child = self._get_child_object(item, key)
        self._children[key] = child
        return child

    def to_dict(self, strip: bool = False):
        return {k: v.to_dict(strip=strip) for k, v in self._children.items()}

    @classmethod
    def from_dict(cls, classes_map: dict, children=None, default_kind=""):
        if children is None:
            return cls(classes_map, default_kind)
        if not isinstance(children, dict):
            raise ValueError("children must be a dict")

        new_obj = cls(classes_map, default_kind)
        for name, child in children.items():
            obj_name = name
            if hasattr(child, "name") and child.name is not None:
                obj_name = child.name
            elif isinstance(child, dict) and "name" in child:
                obj_name = child["name"]
            child_obj = new_obj._get_child_object(child, obj_name)
            new_obj._children[name] = child_obj

        return new_obj

    def _get_child_object(self, child, name):
        if hasattr(child, "kind") and child.kind in self._classes_map.keys():
            child.name = name
            return child
        elif isinstance(child, dict):
            kind = child.get("kind", self._default_kind)
            if kind not in self._classes_map.keys():
                raise ValueError(f"illegal object kind {kind}")
            child_obj = self._classes_map[kind].from_dict(child)
            child_obj.name = name
            return child_obj
        else:
            raise ValueError(f"illegal child (should be dict or child kind), {child}")

    def to_yaml(self):
        return dict_to_yaml(self.to_dict())

    def to_json(self):
        return dict_to_json(self.to_dict())

    def to_str(self):
        return self.__str__()

    def __str__(self):
        return str(self.to_dict())

    def copy(self):
        return deepcopy(self)


class ObjectList:
    def __init__(self, child_class):
        self._children = OrderedDict()
        self._child_class = child_class

    def values(self):
        return self._children.values()

    def keys(self):
        return self._children.keys()

    def items(self):
        return self._children.items()

    def __len__(self):
        return len(self._children)

    def __iter__(self):
        yield from self._children.values()

    def __getitem__(self, name):
        if isinstance(name, int):
            return list(self._children.values())[name]
        return self._children[name]

    def __setitem__(self, key, item):
        self.update(item, key)

    def __delitem__(self, key):
        del self._children[key]

    def to_dict(self, strip: bool = False):
        # method used by ModelObj class to serialize the object to nested dict
        return [t.to_dict(strip=strip) for t in self._children.values()]

    @classmethod
    def from_list(cls, child_class, children=None):
        if children is None:
            return cls(child_class)
        if not isinstance(children, list):
            raise ValueError("states must be a list")

        new_obj = cls(child_class)
        for child in children:
            name, child_obj = new_obj._get_child_object(child)
            new_obj._children[name] = child_obj
        return new_obj

    def _get_child_object(self, child):
        if isinstance(child, self._child_class):
            return child.name, child
        elif isinstance(child, dict):
            if "name" not in child.keys():
                raise ValueError("illegal object no 'name' field")
            child_obj = self._child_class.from_dict(child)
            return child_obj.name, child_obj
        else:
            raise ValueError(f"illegal child (should be dict or child kind), {child}")

    def update(self, child, name=None):
        object_name, child_obj = self._get_child_object(child)
        child_obj.name = name or object_name
        self._children[child_obj.name] = child_obj
        return child_obj

    def move_to_end(self, child, last=True):
        self._children.move_to_end(child, last)

    def update_list(self, object_list: "ObjectList", push_at_start: bool = False):
        if push_at_start:
            self._children = OrderedDict(
                list(object_list._children.items()) + list(self._children.items())
            )
        else:
            self._children = OrderedDict(
                list(self._children.items()) + list(object_list._children.items())
            )


class Credentials(ModelObj):
    generate_access_key = "$generate"
    secret_reference_prefix = "$ref:"

    def __init__(
        self,
        access_key: Optional[str] = None,
    ):
        self.access_key = access_key


class BaseMetadata(ModelObj):
    _default_fields_to_strip = ModelObj._default_fields_to_strip + [
        "hash",
        "uid",
        # Below are environment specific fields, no need to keep when stripping
        "namespace",
        "project",
        "labels",
        "annotations",
        "credentials",
        # Below are state fields, no need to keep when stripping
        "updated",
    ]

    def __init__(
        self,
        name=None,
        tag=None,
        hash=None,
        namespace=None,
        project=None,
        labels=None,
        annotations=None,
        categories=None,
        updated=None,
        credentials=None,
        uid=None,
    ):
        self.name = name
        self.tag = tag
        self.hash = hash
        self.uid = uid
        self.namespace = namespace
        self.project = project or ""
        self.labels = labels or {}
        self.categories = categories or []
        self.annotations = annotations or {}
        self.updated = updated
        self._credentials = None
        self.credentials = credentials

    @property
    def credentials(self) -> Credentials:
        return self._credentials

    @credentials.setter
    def credentials(self, credentials):
        self._credentials = self._verify_dict(credentials, "credentials", Credentials)


class ImageBuilder(ModelObj):
    """An Image builder"""

    def __init__(
        self,
        functionSourceCode=None,  # noqa: N803 - should be "snake_case", kept for BC
        source=None,
        image=None,
        base_image=None,
        commands=None,
        extra=None,
        secret=None,
        code_origin=None,
        registry=None,
        load_source_on_run=None,
        origin_filename=None,
        with_mlrun=None,
        auto_build=None,
        requirements: Optional[list] = None,
        extra_args=None,
        builder_env=None,
        source_code_target_dir=None,
    ):
        self.functionSourceCode = functionSourceCode  #: functionSourceCode
        self.codeEntryType = ""  #: codeEntryType
        self.codeEntryAttributes = ""  #: codeEntryAttributes
        self.source = source  #: source
        self.code_origin = code_origin  #: code_origin
        self.origin_filename = origin_filename
        self.image = image  #: image
        self.base_image = base_image  #: base_image
        self.commands = commands or []  #: commands
        self.extra = extra  #: extra
        self.extra_args = extra_args  #: extra args
        self.builder_env = builder_env  #: builder env
        self.secret = secret  #: secret
        self.registry = registry  #: registry
        self.load_source_on_run = load_source_on_run  #: load_source_on_run
        self.with_mlrun = with_mlrun  #: with_mlrun
        self.auto_build = auto_build  #: auto_build
        self.build_pod = None
        self.requirements = requirements or []  #: pip requirements
        self.source_code_target_dir = source_code_target_dir or None

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if source and not (
            source.startswith("git://")
            # lenient check for file extension because we support many file types locally and remotely
            or pathlib.Path(source).suffix
            or source in [".", "./"]
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"source ({source}) must be a compressed (tar.gz / zip) file, a git repo, "
                f"a file path or in the project's context (.)"
            )

        self._source = source

    def build_config(
        self,
        image="",
        base_image=None,
        commands: Optional[list] = None,
        secret=None,
        source=None,
        extra=None,
        load_source_on_run=None,
        with_mlrun=None,
        auto_build=None,
        requirements=None,
        requirements_file=None,
        overwrite=False,
        builder_env=None,
        extra_args=None,
        source_code_target_dir=None,
    ):
        if image:
            self.image = image
        if base_image:
            self.base_image = base_image
        if commands:
            self.with_commands(commands, overwrite=overwrite)
        if requirements or requirements_file:
            self.with_requirements(requirements, requirements_file, overwrite=overwrite)
        if extra:
            self.extra = extra
        if secret is not None:
            self.secret = secret
        if source:
            self.source = source
        if load_source_on_run:
            self.load_source_on_run = load_source_on_run
        if with_mlrun is not None:
            self.with_mlrun = with_mlrun
        if auto_build:
            self.auto_build = auto_build
        if builder_env:
            self.builder_env = builder_env
        if extra_args:
            self.extra_args = extra_args
        if source_code_target_dir:
            self.source_code_target_dir = source_code_target_dir

    def with_commands(
        self,
        commands: list[str],
        overwrite: bool = False,
    ):
        """add commands to build spec.

        :param commands:  list of commands to run during build
        :param overwrite: whether to overwrite the existing commands or add to them (the default)

        :return: function object
        """
        if not isinstance(commands, list) or not all(
            isinstance(item, str) for item in commands
        ):
            raise ValueError("commands must be a string list")
        if not self.commands or overwrite:
            self.commands = commands
        else:
            # add commands to existing build commands
            for command in commands:
                if command not in self.commands:
                    self.commands.append(command)
            # using list(set(x)) won't retain order,
            # solution inspired from https://stackoverflow.com/a/17016257/8116661
            self.commands = list(dict.fromkeys(self.commands))

    def with_requirements(
        self,
        requirements: Optional[list[str]] = None,
        requirements_file: str = "",
        overwrite: bool = False,
    ):
        """add package requirements from file or list to build spec.

        :param requirements:        a list of python packages
        :param requirements_file:   path to a python requirements file
        :param overwrite:           overwrite existing requirements,
                                    when False (default) will append to existing requirements
        :return: function object
        """
        requirements = requirements or []
        self._verify_list(requirements, "requirements")
        resolved_requirements = self.resolve_requirements(
            requirements, requirements_file
        )
        requirements = self.requirements or [] if not overwrite else []

        # make sure we don't append the same line twice
        for requirement in resolved_requirements:
            if requirement not in requirements:
                requirements.append(requirement)

        self.requirements = requirements

    @staticmethod
    def resolve_requirements(requirements: list, requirements_file: str = "") -> list:
        requirements = requirements or []
        requirements_to_resolve = []

        # handle the requirements_file argument
        if requirements_file:
            with open(requirements_file) as fp:
                requirements_to_resolve.extend(fp.read().splitlines())

        # handle the requirements argument
        requirements_to_resolve.extend(requirements)

        requirements = []
        for requirement in requirements_to_resolve:
            # clean redundant leading and trailing whitespaces
            requirement = requirement.strip()

            # ignore empty lines
            # ignore comments
            if not requirement or requirement.startswith("#"):
                continue

            # ignore inline comments as well
            inline_comment = requirement.split(" #")
            if len(inline_comment) > 1:
                requirement = inline_comment[0].strip()

            requirements.append(requirement)

        return requirements


class Notification(ModelObj):
    """Notification object

    :param kind: notification implementation kind - slack, webhook, etc. See
        :py:class:`mlrun.common.schemas.notification.NotificationKind`
    :param name: for logging and identification
    :param message: message content in the notification
    :param severity: severity to display in the notification
    :param when: list of statuses to trigger the notification: 'running', 'completed', 'error'
    :param condition: optional condition to trigger the notification, a jinja2 expression that can use run data
                      to evaluate if the notification should be sent in addition to the 'when' statuses.
                      e.g.: '{{ run["status"]["results"]["accuracy"] < 0.9}}'
    :param params: Implementation specific parameters for the notification implementation (e.g. slack webhook url,
                   git repository details, etc.)
    :param secret_params: secret parameters for the notification implementation, same as params but will be stored
                          in a k8s secret and passed as a secret reference to the implementation.
    :param status: notification status - pending, sent, error
    :param sent_time: time the notification was sent
    :param reason: failure reason if the notification failed to send
    """

    def __init__(
        self,
        kind: mlrun.common.schemas.notification.NotificationKind = (
            mlrun.common.schemas.notification.NotificationKind.slack
        ),
        name=None,
        message=None,
        severity: mlrun.common.schemas.notification.NotificationSeverity = (
            mlrun.common.schemas.notification.NotificationSeverity.INFO
        ),
        when=None,
        condition=None,
        secret_params=None,
        params=None,
        status=None,
        sent_time=None,
        reason=None,
    ):
        self.kind = kind
        self.name = name or ""
        self.message = message or ""
        self.severity = severity
        self.when = when or ["completed"]
        self.condition = condition or ""
        self.secret_params = secret_params or {}
        self.params = params or {}
        self.status = status
        self.sent_time = sent_time
        self.reason = reason

        self.validate_notification()

    def validate_notification(self):
        try:
            mlrun.common.schemas.notification.Notification(**self.to_dict())
        except pydantic.v1.error_wrappers.ValidationError as exc:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Invalid notification object"
            ) from exc

        # validate that size of notification secret_params doesn't exceed 1 MB,
        # due to k8s default secret size limitation.
        # a buffer of 100 KB is added to the size to account for the size of the secret metadata
        if (
            len(json.dumps(self.secret_params))
            > mlrun.common.schemas.notification.NotificationLimits.max_params_size.value
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Notification params size exceeds max size of 1 MB"
            )

    def validate_notification_params(self, default_notification_params=None):
        default_notification_params = default_notification_params or {}
        notification_type = mlrun.utils.notifications.NotificationTypes(self.kind)
        notification_class = notification_type.get_notification()
        secret_params = self.secret_params or {}
        params = self.params or {}
        default_params = default_notification_params.get(notification_type, {})
        params = notification_class.enrich_default_params(params, default_params)
        # if the secret_params are already masked - no need to validate
        params_secret = secret_params.get("secret", "")
        if params_secret:
            if len(secret_params) > 1:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "When the 'secret' key is present, 'secret_params' should not contain any other keys."
                )
            return

        if not secret_params and not params:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Both 'secret_params' and 'params' are empty, at least one must be defined."
            )

        notification_class.validate_params(secret_params | params)

    def enrich_unmasked_secret_params_from_project_secret(self):
        """
        Fill the notification secret params from the project secret.
        We are using this function instead of unmask_secret_params_from_project_secret when we run inside the
        workflow runner pod that doesn't have access to the k8s secrets (but have access to the project secret)
        """
        secret = self.secret_params.get("secret")
        if secret:
            secret_value = mlrun.get_secret_or_env(secret)
            if secret_value:
                try:
                    self.secret_params = json.loads(secret_value)
                except ValueError as exc:
                    raise mlrun.errors.MLRunValueError(
                        "Failed to parse secret value"
                    ) from exc

    @staticmethod
    def validate_notification_uniqueness(notifications: list["Notification"]):
        """Validate that all notifications in the list are unique by name"""
        names = [notification.name for notification in notifications]
        if len(names) != len(set(names)):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Notification names must be unique"
            )


class RunMetadata(ModelObj):
    """Run metadata"""

    def __init__(
        self,
        uid=None,
        name=None,
        project=None,
        labels=None,
        annotations=None,
        iteration=None,
    ):
        self.uid = uid
        self._iteration = iteration
        self.name = name
        self.project = project
        self.labels = labels or {}
        self.annotations = annotations or {}

    @property
    def iteration(self):
        return self._iteration or 0

    @iteration.setter
    def iteration(self, iteration):
        self._iteration = iteration

    def is_workflow_runner(self):
        if not self.labels:
            return False
        return (
            self.labels.get(mlrun_constants.MLRunInternalLabels.job_type, "")
            == "workflow-runner"
        )


class HyperParamStrategies:
    grid = "grid"
    list = "list"
    random = "random"
    custom = "custom"

    @staticmethod
    def all():
        return [
            HyperParamStrategies.grid,
            HyperParamStrategies.list,
            HyperParamStrategies.random,
            HyperParamStrategies.custom,
        ]


class HyperParamOptions(ModelObj):
    """Hyper Parameter Options

    Parameters:
        param_file (str):                   hyper params input file path/url, instead of inline
        strategy (HyperParamStrategies):    hyper param strategy - grid, list or random
        selector (str):                     selection criteria for best result ([min|max.]<result>), e.g. max.accuracy
        stop_condition (str):               early stop condition e.g. "accuracy > 0.9"
        parallel_runs (int):                number of param combinations to run in parallel (over Dask)
        dask_cluster_uri (str):             db uri for a deployed dask cluster function, e.g. db://myproject/dask
        max_iterations (int):               max number of runs (in random strategy)
        max_errors (int):                   max number of child runs errors for the overall job to fail
        teardown_dask (bool):               kill the dask cluster pods after the runs
    """

    def __init__(
        self,
        param_file=None,
        strategy: typing.Optional[HyperParamStrategies] = None,
        selector=None,
        stop_condition=None,
        parallel_runs=None,
        dask_cluster_uri=None,
        max_iterations=None,
        max_errors=None,
        teardown_dask=None,
    ):
        self.param_file = param_file
        self.strategy = strategy
        self.selector = selector
        self.stop_condition = stop_condition
        self.max_iterations = max_iterations
        self.max_errors = max_errors
        self.parallel_runs = parallel_runs
        self.dask_cluster_uri = dask_cluster_uri
        self.teardown_dask = teardown_dask

    def validate(self):
        if self.strategy and self.strategy not in HyperParamStrategies.all():
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"illegal hyper param strategy, use {','.join(HyperParamStrategies.all())}"
            )
        if self.max_iterations and self.strategy != HyperParamStrategies.random:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "max_iterations is only valid in random strategy"
            )


class RetryBackoff(ModelObj):
    """Backoff strategy for retries."""

    def __init__(self, base_delay: Optional[str] = None):
        # The base_delay time string must conform to timelength python package standards and be at least
        # mlrun.mlconf.function.spec.retry.backoff.min_base_delay (e.g. 1000s, 1 hour 30m, 1h etc.).
        self.base_delay = (
            base_delay or mlrun.mlconf.function.spec.retry.backoff.default_base_delay
        )


class Retry(ModelObj):
    """Retry configuration"""

    def __init__(
        self,
        count: int = 0,
        backoff: typing.Union[RetryBackoff, dict] = None,
    ):
        # Set to None if count is 0 to eliminate the retry configuration from the dictionary representation.
        self.count = count or None
        self.backoff = backoff

    @property
    def backoff(self) -> Optional[RetryBackoff]:
        if not self.count:
            # Retry is not configured, return None
            return None
        return self._backoff

    @backoff.setter
    def backoff(self, backoff):
        self._backoff = self._verify_dict(backoff, "backoff", RetryBackoff)


class RunSpec(ModelObj):
    """Run specification"""

    _fields_to_serialize = ModelObj._fields_to_serialize + [
        "handler",
        "affinity",
        "tolerations",
    ]

    def __init__(
        self,
        parameters=None,
        hyperparams=None,
        param_file=None,
        selector=None,
        handler=None,
        inputs=None,
        outputs=None,
        input_path=None,
        output_path=None,
        function=None,
        secret_sources=None,
        data_stores=None,
        strategy=None,
        verbose=None,
        scrape_metrics=None,
        hyper_param_options=None,
        allow_empty_resources=None,
        inputs_type_hints=None,
        returns=None,
        notifications=None,
        state_thresholds=None,
        reset_on_run=None,
        node_selector=None,
        tolerations=None,
        affinity=None,
        retry=None,
    ):
        # A dictionary of parsing configurations that will be read from the inputs the user set. The keys are the inputs
        # keys (parameter names) and the values are the type hint given in the input keys after the colon.
        # Notice: We set it first as empty dictionary as setting the inputs will set it as well in case the type hints
        # were passed in the input keys.
        self._inputs_type_hints = {}

        self._hyper_param_options = None

        # Initialize the inputs and returns properties first and then use their setter methods:
        self._inputs = None
        self.inputs = inputs
        if inputs_type_hints:
            # Override the empty dictionary only if the user passed the parameter:
            self._inputs_type_hints = inputs_type_hints
        self._returns = None
        self.returns = returns

        self._outputs = outputs
        self.hyper_param_options = hyper_param_options
        self.parameters = parameters or {}
        self.hyperparams = hyperparams or {}
        self.param_file = param_file
        self.strategy = strategy
        self.selector = selector
        self.handler = handler
        self.input_path = input_path
        self.output_path = output_path
        self.function = function
        self._secret_sources = secret_sources or []
        self._data_stores = data_stores
        self.verbose = verbose
        self.scrape_metrics = scrape_metrics
        self.allow_empty_resources = allow_empty_resources
        self._notifications = notifications or []
        self.state_thresholds = state_thresholds or {}
        self.reset_on_run = reset_on_run
        self.node_selector = node_selector or {}
        self.tolerations = tolerations or {}
        self.affinity = affinity or {}
        self.retry = retry or {}

    def _serialize_field(
        self, struct: dict, field_name: Optional[str] = None, strip: bool = False
    ) -> Optional[str]:
        # We pull the field from self and not from struct because it was excluded from the struct
        if field_name == "handler":
            if self.handler and isinstance(self.handler, str):
                return self.handler
            return None

        # Properly serialize known K8s objects
        if field_name in {"affinity", "tolerations"}:
            value = getattr(self, field_name, None)
            if hasattr(value, "to_dict"):
                return value.to_dict()
            return value

        return super()._serialize_field(struct, field_name, strip)

    def is_hyper_job(self):
        param_file = self.param_file or self.hyper_param_options.param_file
        return param_file or self.hyperparams

    @property
    def inputs(self) -> dict[str, str]:
        """
        Get the inputs dictionary. A dictionary of parameter names as keys and paths as values.

        :return: The inputs dictionary.
        """
        return self._inputs

    @inputs.setter
    def inputs(self, inputs: dict[str, str]):
        """
        Set the given inputs in the spec. Inputs can include a type hint string in their keys following a colon, meaning
        following this structure: "<input key : type hint>".

        :exmaple:

        >>> run_spec.inputs = {
        ...     "my_input": "...",
        ...     "my_hinted_input : pandas.DataFrame": "...",
        ... }

        :param inputs: The inputs to set.
        """
        # Check if None, then set and return:
        if inputs is None:
            self._inputs = None
            return

        # Verify it's a dictionary:
        self._inputs = self._verify_dict(inputs, "inputs")

    @property
    def inputs_type_hints(self) -> dict[str, str]:
        """
        Get the input type hints. A dictionary of parameter names as keys and their type hints as values.

        :return: The input type hints dictionary.
        """
        return self._inputs_type_hints

    @inputs_type_hints.setter
    def inputs_type_hints(self, inputs_type_hints: dict[str, str]):
        """
        Set the inputs type hints to parse during a run.

        :param inputs_type_hints: The type hints to set.
        """
        # Verify the given value is a dictionary or None:
        self._inputs_type_hints = self._verify_dict(
            inputs_type_hints, "inputs_type_hints"
        )

    @property
    def returns(self):
        """
        Get the returns list. A list of log hints for returning values.

        :return: The returns list.
        """
        return self._returns

    @returns.setter
    def returns(self, returns: list[Union[str, dict[str, str]]]):
        """
        Set the returns list to log the returning values at the end of a run.

        :param returns: The return list to set.

        :raise MLRunInvalidArgumentError: In case one of the values in the list is invalid.
        """
        # This import is located in the method due to circular imports error.
        from mlrun.package.utils import LogHintUtils

        if returns is None:
            self._returns = None
            return
        self._verify_list(returns, "returns")

        # Validate:
        for log_hint in returns:
            LogHintUtils.parse_log_hint(log_hint=log_hint)

        # Store the results:
        self._returns = returns

    @property
    def hyper_param_options(self) -> HyperParamOptions:
        return self._hyper_param_options

    @hyper_param_options.setter
    def hyper_param_options(self, hyper_param_options):
        self._hyper_param_options = self._verify_dict(
            hyper_param_options, "hyper_param_options", HyperParamOptions
        )

    @property
    def outputs(self) -> list[str]:
        """
        Get the expected outputs. The list is constructed from keys of both the `outputs` and `returns` properties.

        :return: The expected outputs list.
        """
        return self.join_outputs_and_returns(
            outputs=self._outputs, returns=self.returns
        )

    @outputs.setter
    def outputs(self, outputs):
        """
        Set the expected outputs list.

        :param outputs: A list of expected output keys.
        """
        self._verify_list(outputs, "outputs")
        self._outputs = outputs

    @property
    def secret_sources(self):
        return self._secret_sources

    @secret_sources.setter
    def secret_sources(self, secret_sources):
        self._verify_list(secret_sources, "secret_sources")
        self._secret_sources = secret_sources

    @property
    def data_stores(self):
        return self._data_stores

    @data_stores.setter
    def data_stores(self, data_stores):
        self._verify_list(data_stores, "data_stores")
        self._data_stores = data_stores

    @property
    def handler_name(self):
        if self.handler:
            if inspect.isfunction(self.handler):
                return self.handler.__name__
            else:
                return str(self.handler)
        return ""

    @property
    def notifications(self):
        return self._notifications

    @notifications.setter
    def notifications(self, notifications):
        if isinstance(notifications, list):
            self._notifications = ObjectList.from_list(Notification, notifications)
        elif isinstance(notifications, ObjectList):
            self._notifications = notifications
        else:
            raise ValueError("Notifications must be a list")

    @property
    def state_thresholds(self):
        return self._state_thresholds

    @state_thresholds.setter
    def state_thresholds(self, state_thresholds: dict[str, str]):
        """
        Set the dictionary of k8s resource states to thresholds time strings.
        The state will be matched against the pod's status. The threshold should be a time string that conforms
        to timelength python package standards and is at least 1 minute (-1 for infinite). If the phase is active
        for longer than the threshold, the run will be marked as aborted and the pod will be deleted.
        See mlconf.function.spec.state_thresholds for the state options and default values.

        example:
            {"image_pull_backoff": "1h", "executing": "1d 2 hours"}

        :param state_thresholds: The state-thresholds dictionary.
        """
        self._verify_dict(state_thresholds, "state_thresholds")
        self._state_thresholds = state_thresholds

    @property
    def retry(self) -> Retry:
        return self._retry

    @retry.setter
    def retry(self, retry: typing.Union[Retry, dict]):
        self._retry = self._verify_dict(retry, "retry", Retry)

    def extract_type_hints_from_inputs(self):
        """
        This method extracts the type hints from the input keys in the input dictionary.

        As a result, after the method ran the inputs dictionary - a dictionary of parameter names as keys and paths as
        values, will be cleared from type hints and the extracted type hints will be saved in the spec's inputs type
        hints dictionary - a dictionary of parameter names as keys and their type hints as values. If a parameter is
        not in the type hints dictionary, its type hint will be `mlrun.DataItem` by default.
        """
        # Validate there are inputs to read:
        if self.inputs is None:
            return

        # Prepare dictionaries to hold the cleared inputs and type hints:
        cleared_inputs = {}
        extracted_inputs_type_hints = {}

        # Clear the inputs from parsing configurations:
        for input_key, input_value in self.inputs.items():
            # Look for type hinted in input key:
            if ":" in input_key:
                # Separate the user input by colon:
                input_key, input_type = RunSpec._separate_type_hint_from_input_key(
                    input_key=input_key
                )
                # Collect the type hint:
                extracted_inputs_type_hints[input_key] = input_type
            # Collect the cleared input key:
            cleared_inputs[input_key] = input_value

        # Set the now configuration free inputs and extracted type hints:
        self.inputs = cleared_inputs
        self.inputs_type_hints = extracted_inputs_type_hints

    @staticmethod
    def join_outputs_and_returns(
        outputs: list[str], returns: list[Union[str, dict[str, str]]]
    ) -> list[str]:
        """
        Get the outputs set in the spec. The outputs are constructed from both the 'outputs' and 'returns' properties
        that were set by the user.

        :param outputs: A spec outputs property - list of output keys.
        :param returns: A spec returns property - list of key and configuration of how to log returning values.

        :return: The joined 'outputs' and 'returns' list.
        """
        # Collect the 'returns' property keys:
        cleared_returns = []
        if returns:
            for return_value in returns:
                # Check if the return entry is a configuration dictionary or a key-type structure string (otherwise its
                # just a key string):
                if isinstance(return_value, dict):
                    # Set it to the artifact key:
                    return_value = return_value["key"]
                elif ":" in return_value:
                    # Take only the key name (returns values pattern is validated when set in the spec):
                    return_value = return_value.replace(" ", "").split(":")[0]
                # Collect it:
                cleared_returns.append(return_value)

        # Use `set` join to combine the two lists without duplicates:
        outputs = list(set(outputs if outputs else []) | set(cleared_returns))

        return outputs

    @staticmethod
    def _separate_type_hint_from_input_key(input_key: str) -> tuple[str, str]:
        """
        An input key in the `inputs` dictionary parameter of a task (or `Runtime.run` method) or the docs setting of a
        `Runtime` handler can be provided with a colon to specify its type hint in the following structure:
        "<parameter_key> : <type_hint>".

        This method parses the provided value by the user.

        :param input_key: A string entry in the inputs dictionary keys.

        :return: The value as key and type hint tuple.

        :raise MLRunInvalidArgumentError: If an incorrect pattern was provided.
        """
        # Validate correct pattern:
        if input_key.count(":") > 1:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Incorrect input pattern. Input keys can have only a single ':' in them to specify the desired type "
                f"the input will be parsed as. Given: {input_key}."
            )

        # Split into key and type:
        value_key, value_type = input_key.replace(" ", "").split(":")

        return value_key, value_type


class RunStatus(ModelObj):
    """Run status"""

    _default_fields_to_strip = ModelObj._default_fields_to_strip + ["artifacts"]

    def __init__(
        self,
        state=None,
        error=None,
        host=None,
        commit=None,
        status_text=None,
        results=None,
        artifacts=None,
        start_time=None,
        end_time=None,
        last_update=None,
        iterations=None,
        ui_url=None,
        reason: Optional[str] = None,
        notifications: Optional[dict[str, Notification]] = None,
        artifact_uris: Optional[dict[str, str]] = None,
        retry_count: Optional[int] = None,
        retries: Optional[list[dict]] = None,
    ):
        self.state = state or "created"
        self.status_text = status_text
        self.error = error
        self.host = host
        self.commit = commit
        self.results = results
        self._artifacts = artifacts
        self.start_time = start_time
        self.end_time = end_time
        self.last_update = last_update
        self.iterations = iterations
        self.ui_url = ui_url
        self.reason = reason
        self.notifications = notifications or {}
        # Artifact key -> URI mapping, since the full artifacts are not stored in the runs DB table
        self._artifact_uris = artifact_uris or {}
        self._retry_count = retry_count or None
        self._retries = retries or []

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: Optional[dict] = None
    ):
        deprecated_fields = {
            # Set artifacts as deprecated for lazy loading
            "artifacts": "artifact_uris"
        }
        return super().from_dict(
            struct, fields=fields, deprecated_fields=deprecated_fields
        )

    @property
    def artifacts(self):
        """
        Artifacts are lazy loaded to reduce memory consumption.
        We keep artifact_uris (key -> store URI dictionary) to be able to get the run artifacts easily.
        If the artifact is not already in the cache, we get it from the store (DB).
        :return: List of artifact dictionaries
        """
        self._artifacts = self._artifacts or []
        existing_artifact_keys = {
            artifact["metadata"]["key"] for artifact in self._artifacts
        }
        for key, uri in self.artifact_uris.items():
            if key not in existing_artifact_keys:
                artifact = mlrun.datastore.get_store_resource(uri)
                self._artifacts.append(artifact.to_dict())
        return self._artifacts

    @artifacts.setter
    def artifacts(self, artifacts):
        self._artifacts = artifacts

    @property
    def artifact_uris(self):
        return self._artifact_uris

    @artifact_uris.setter
    def artifact_uris(self, artifact_uris):
        resolved_artifact_uris = {}
        if isinstance(artifact_uris, list):
            # artifact_uris is the deprecated list of artifacts - convert to new form
            for artifact in artifact_uris:
                if isinstance(artifact, dict):
                    artifact = mlrun.artifacts.dict_to_artifact(artifact)
                resolved_artifact_uris[artifact.key] = artifact.uri
        else:
            resolved_artifact_uris = artifact_uris

        self._artifact_uris = resolved_artifact_uris

    @property
    def retry_count(self) -> Optional[int]:
        """
        The number of retries that were made for this run.
        """
        return self._retry_count

    @retry_count.setter
    def retry_count(self, retry_count: int):
        """
        Set the number of retries that were made for this run.
        :param retry_count: The number of retries.
        """
        self._retry_count = retry_count

    @property
    def retries(self) -> list[dict]:
        """List of metadata for each retry attempt."""
        return self._retries

    @retries.setter
    def retries(self, retries: list[dict]):
        """
        Set the list of retry attempt metadata.
        :param retries: A list of dictionaries, each representing a retry attempt.
        """
        self._retries = retries

    def is_failed(self) -> Optional[bool]:
        """
        This method returns whether a run has failed.
        Returns none if state has yet to be defined. callee is responsible for handling None.
        (e.g wait for state to be defined)
        """
        if not self.state:
            return None
        return self.state.casefold() in [
            mlrun.run.RunStatuses.failed.casefold(),
            mlrun.run.RunStatuses.error.casefold(),
        ]


class RunTemplate(ModelObj):
    """Run template"""

    def __init__(self, spec: RunSpec = None, metadata: RunMetadata = None):
        self._spec = None
        self._metadata = None
        self.spec = spec
        self.metadata = metadata

    @property
    def spec(self) -> RunSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", RunSpec)

    @property
    def metadata(self) -> RunMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", RunMetadata)

    def with_params(self, **kwargs):
        """set task parameters using key=value, key2=value2, .."""
        self.spec.parameters = kwargs
        return self

    def with_input(self, key, path):
        """set task data input, path is an Mlrun global DataItem uri

        examples::

            task.with_input("data", "/file-dir/path/to/file")
            task.with_input("data", "s3://<bucket>/path/to/file")
            task.with_input("data", "v3io://<data-container>/path/to/file")
        """
        if not self.spec.inputs:
            self.spec.inputs = {}
        self.spec.inputs[key] = path
        return self

    def with_hyper_params(
        self,
        hyperparams,
        selector=None,
        strategy: HyperParamStrategies = None,
        **options,
    ):
        """set hyper param values and configurations,
        see parameters in: :py:class:`HyperParamOptions`

        example::

            grid_params = {"p1": [2, 4, 1], "p2": [10, 20]}
            task = mlrun.new_task("grid-search")
            task.with_hyper_params(grid_params, selector="max.accuracy")
        """
        self.spec.hyperparams = hyperparams
        self.spec.hyper_param_options = options
        self.spec.hyper_param_options.selector = selector
        self.spec.hyper_param_options.strategy = strategy
        self.spec.hyper_param_options.validate()
        return self

    def with_param_file(
        self,
        param_file,
        selector=None,
        strategy: HyperParamStrategies = None,
        **options,
    ):
        """set hyper param values (from a file url) and configurations,
        see parameters in: :py:class:`HyperParamOptions`

        example::

            grid_params = "s3://<my-bucket>/path/to/params.json"
            task = mlrun.new_task("grid-search")
            task.with_param_file(grid_params, selector="max.accuracy")
        """
        self.spec.hyper_param_options = options
        self.spec.hyper_param_options.param_file = param_file
        self.spec.hyper_param_options.selector = selector
        self.spec.hyper_param_options.strategy = strategy
        self.spec.hyper_param_options.validate()
        return self

    def with_secrets(self, kind, source):
        """register a secrets source (file, env or dict)

        read secrets from a source provider to be used in workflows, example::

            task.with_secrets('file', 'file.txt')
            task.with_secrets('inline', {'key': 'val'})
            task.with_secrets('env', 'ENV1,ENV2')

            task.with_secrets('vault', ['secret1', 'secret2'...])

            # If using with k8s secrets, the k8s secret is managed by MLRun, through the project-secrets
            # mechanism. The secrets will be attached to the running pod as environment variables.
            task.with_secrets('kubernetes', ['secret1', 'secret2'])

            # If using an empty secrets list [] then all accessible secrets will be available.
            task.with_secrets('vault', [])

            # To use with Azure key vault, a k8s secret must be created with the following keys:
            # kubectl -n <namespace> create secret generic azure-key-vault-secret \\
            #     --from-literal=tenant_id=<service principal tenant ID> \\
            #     --from-literal=client_id=<service principal client ID> \\
            #     --from-literal=secret=<service principal secret key>

            task.with_secrets('azure_vault', {
                'name': 'my-vault-name',
                'k8s_secret': 'azure-key-vault-secret',
                # An empty secrets list may be passed ('secrets': []) to access all vault secrets.
                'secrets': ['secret1', 'secret2'...]
            })

        :param kind:   secret type (file, inline, env)
        :param source: secret data or link (see example)

        :returns: The RunTemplate object
        """
        if kind == "azure_vault" and isinstance(source, dict):
            candidate_secret_name = (source.get("k8s_secret") or "").strip()
            if candidate_secret_name:
                mlrun.common.secrets.validate_not_forbidden_secret(
                    candidate_secret_name
                )
        if kind == "vault" and isinstance(source, list):
            source = {"project": self.metadata.project, "secrets": source}

        self.spec.secret_sources.append({"kind": kind, "source": source})
        return self

    def set_label(self, key, value):
        """set a key/value label for the task"""
        self.metadata.labels[key] = str(value)
        return self

    def to_env(self):
        environ["MLRUN_EXEC_CONFIG"] = self.to_json()


class RunObject(RunTemplate):
    """A run"""

    def __init__(
        self,
        spec: RunSpec = None,
        metadata: RunMetadata = None,
        status: RunStatus = None,
    ):
        super().__init__(spec, metadata)
        self._status = None
        self.status = status
        self.outputs_wait_for_completion = True

    @classmethod
    def from_template(cls, template: RunTemplate):
        return cls(template.spec, template.metadata)

    def to_json(self, exclude=None, **kwargs):
        # Since the `params` attribute within each notification object can be large,
        # it has the potential to cause errors and is unnecessary for the notification functionality.
        # Therefore, in this section, we remove the `params` attribute from each notification object.
        if (
            exclude_notifications_params := kwargs.get("exclude_notifications_params")
        ) and exclude_notifications_params:
            if self.spec.notifications:
                # Extract and remove 'params' from each notification
                extracted_params = []
                for notification in self.spec.notifications:
                    extracted_params.append(notification.params)
                    del notification.params
                # Generate the JSON representation, excluding specified fields
                json_obj = super().to_json(exclude=exclude)
                # Restore 'params' back to the notifications
                for notification, params in zip(
                    self.spec.notifications, extracted_params
                ):
                    notification.params = params
                return json_obj
        return super().to_json(exclude=exclude)

    @property
    def status(self) -> RunStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", RunStatus)

    @property
    def error(self) -> str:
        """error string if failed"""
        if (
            self.status
            and self.status.state
            in mlrun.common.runtimes.constants.RunStates.error_and_abortion_states()
        ):
            unknown_error = ""
            if (
                self.status.state
                in mlrun.common.runtimes.constants.RunStates.abortion_states()
            ):
                unknown_error = "Run was aborted"

            elif (
                self.status.state
                in mlrun.common.runtimes.constants.RunStates.error_states()
            ):
                unknown_error = "Unknown error"

            return (
                self.status.error
                or self.status.status_text
                or self.status.reason
                or unknown_error
            )
        return ""

    def output(self, key: str):
        """
        Return the value of a specific result or artifact by key.

        This method waits for the outputs to complete and retrieves the value corresponding to the provided key.
        If the key exists in the results, it returns the corresponding result value.
        If not found in results, it attempts to fetch the artifact by key (cached in the run status).
        If the artifact is not found, it tries to fetch the artifact URI by key.
        If no artifact or result is found for the key, returns None.

        :param key: The key of the result or artifact to retrieve.
        :return: The value of the result or the artifact URI corresponding to the key, or None if not found.
        """
        self._outputs_wait_for_completion()

        # Check if the key exists in results and return the result value
        if self.status.results and key in self.status.results:
            return self.status.results[key]

        # Artifacts are usually cached in the run object under `status.artifacts`. However, the artifacts are not
        # stored in the DB as part of the run. The server may enrich the run with the artifacts or provide
        # `status.artifact_uris` instead. See mlrun.common.formatters.run.RunFormat.
        # When running locally - `status.artifact_uri` does not exist in the run.
        # When listing runs - `status.artifacts` does not exist in the run.
        artifact = self._artifact(key)
        if artifact:
            return get_artifact_target(artifact, self.metadata.project)

        if self.status.artifact_uris and key in self.status.artifact_uris:
            return self.status.artifact_uris[key]

        return None

    @property
    def ui_url(self) -> str:
        """UI URL (for relevant runtimes)"""
        self.refresh()
        if not self._status.ui_url:
            print(f"UI currently not available (status={self._status.state})")
        return self._status.ui_url

    @property
    def outputs(self):
        """
        Return a dictionary of outputs, including result values and artifact URIs.

        This method waits for the outputs to complete and combines result values
        and artifact URIs into a single dictionary. If there are multiple artifacts
        for the same key, only include the artifact that does not have the "latest" tag.
        If there is no other tag, include the "latest" tag as a fallback.

        :return: Dictionary containing result values and artifact URIs.
        """
        self._outputs_wait_for_completion()
        outputs = {}

        # Add results if available
        if self.status.results:
            outputs.update(self.status.results)

        # Artifacts are usually cached in the run object under `status.artifacts`. However, the artifacts are not
        # stored in the DB as part of the run. The server may enrich the run with the artifacts or provide
        # `status.artifact_uris` instead. See mlrun.common.formatters.run.RunFormat.
        # When running locally - `status.artifact_uri` does not exist in the run.
        # When listing runs - `status.artifacts` does not exist in the run.
        if self.status.artifacts:
            outputs.update(self._process_artifacts(self.status.artifacts))
        elif self.status.artifact_uris:
            outputs.update(self.status.artifact_uris)

        return outputs

    def artifact(self, key: str) -> typing.Optional["mlrun.DataItem"]:
        """Return artifact DataItem by key.

        This method waits for the outputs to complete, searches for the artifact matching the given key,
        and returns a DataItem if the artifact is found.

        :param key: The key of the artifact to find.
        :return: A DataItem corresponding to the artifact with the given key, or None if no such artifact is found.
        """
        self._outputs_wait_for_completion()
        artifact = self._artifact(key)
        if not artifact:
            return None
        uri = get_artifact_target(artifact, self.metadata.project)
        return mlrun.get_dataitem(uri) if uri else None

    def _outputs_wait_for_completion(
        self,
        show_logs=False,
    ):
        """
        Wait for the run to complete fetching the run outputs.
        When running a function with watch=False, and passing the outputs to another function,
        the outputs will not be available until the run is completed.
        :param show_logs: default False, avoid spamming unwanted logs of the run when the user asks for outputs
        """
        if self.outputs_wait_for_completion:
            self.wait_for_completion(
                show_logs=show_logs,
            )

    def _artifact(self, key):
        """
        Return the last artifact DataItem that matches the given key.

        If multiple artifacts with the same key exist, return the last one in the list.
        If there are artifacts with different tags, the method will return the one with a tag other than 'latest'
        if available.
        If no artifact with the given key is found, return None.

        :param key: The key of the artifact to retrieve.
        :return: The last artifact DataItem with the given key, or None if no such artifact is found.
        """
        if not self.status.artifacts and not self.status.artifact_uris:
            return None

        # Collect artifacts that match the key
        matching_artifacts = [
            artifact
            for artifact in self.status.artifacts
            if artifact["metadata"].get("key") == key
        ]

        if not matching_artifacts:
            if key not in self.status.artifact_uris:
                return None

            # Get artifact by store URI sanity (should have been enriched by now in status.artifacts property)
            artifact_uri = self.status.artifact_uris[key]
            return mlrun.datastore.get_store_resource(artifact_uri)

        # Sort matching artifacts by creation date in ascending order.
        # The last element in the list will be the one created most recently.
        # In case the `created` field does not exist in the artifact, that artifact will appear first in the sorted list
        matching_artifacts.sort(
            key=lambda artifact: artifact["metadata"].get("created", datetime.min)
        )

        # Filter out artifacts with 'latest' tag
        non_latest_artifacts = [
            artifact
            for artifact in matching_artifacts
            if artifact["metadata"].get("tag") != "latest"
        ]

        # Return the last non-'latest' artifact if available, otherwise return the last artifact
        # In the case of only one tag, `status.artifacts` includes [v1, latest]. In that case, we want to return v1.
        # In the case of multiple tags, `status.artifacts` includes [v1, latest, v2, v3].
        # In that case, we need to return the last one (v3).
        return (non_latest_artifacts or matching_artifacts)[-1]

    def _process_artifacts(self, artifacts):
        artifacts_by_key = {}

        # Organize artifacts by key
        for artifact in artifacts:
            key = artifact["metadata"]["key"]
            if key not in artifacts_by_key:
                artifacts_by_key[key] = []
            artifacts_by_key[key].append(artifact)

        outputs = {}
        for key, artifacts in artifacts_by_key.items():
            # Sort matching artifacts by creation date in ascending order.
            # The last element in the list will be the one created most recently.
            # In case the `created` field does not exist in the artifactthat artifact will appear
            # first in the sorted list
            artifacts.sort(
                key=lambda artifact: artifact["metadata"].get("created", datetime.min)
            )

            # Filter out artifacts with 'latest' tag
            non_latest_artifacts = [
                artifact
                for artifact in artifacts
                if artifact["metadata"].get("tag") != "latest"
            ]

            # Save the last non-'latest' artifact if available, otherwise save the last artifact
            # In the case of only one tag, `artifacts` includes [v1, latest], in that case, we want to save v1.
            # In the case of multiple tags, `artifacts` includes [v1, latest, v2, v3].
            # In that case, we need to save the last one (v3).
            artifact_to_save = (non_latest_artifacts or artifacts)[-1]
            outputs[key] = get_artifact_target(artifact_to_save, self.metadata.project)

        return outputs

    def uid(self):
        """run unique id"""
        return self.metadata.uid

    def state(self):
        """current run state"""
        if (
            self.status.state
            in mlrun.common.runtimes.constants.RunStates.terminal_states()
        ):
            return self.status.state
        self.refresh()
        return self.status.state or "unknown"

    def refresh(self):
        """refresh run state from the db"""
        db = mlrun.get_run_db()
        run = db.read_run(
            uid=self.metadata.uid,
            project=self.metadata.project,
            iter=self.metadata.iteration,
        )
        if run:
            run_status = run.get("status", {})
            # Artifacts are not stored in the DB, so we need to preserve them here
            run_status["artifacts"] = self.status.artifacts
            self.status = RunStatus.from_dict(run_status)
            return self

    def show(self):
        """show the current status widget, in jupyter notebook"""
        db = mlrun.get_run_db()
        db.list_runs(uid=self.metadata.uid, project=self.metadata.project).show()

    def logs(self, watch=True, db=None, offset=0):
        """return or watch on the run logs"""
        if not db:
            db = mlrun.get_run_db()

        if not db:
            logger.warning("DB is not configured, cannot show logs")
            return None

        state, new_offset = db.watch_log(
            self.metadata.uid, self.metadata.project, watch=watch, offset=offset
        )
        if state:
            logger.debug("Run reached terminal state", state=state)

        return state, new_offset

    def wait_for_completion(
        self,
        sleep=3,
        timeout=0,
        raise_on_failure=True,
        show_logs=None,
        logs_interval=None,
    ):
        """
        Wait for remote run to complete.
        Default behavior is to wait until reached terminal state or timeout passed, if timeout is 0 then wait forever
        It pulls the run status from the db every sleep seconds.
        If show_logs is not False and logs_interval is not None, it will print the logs when run reached terminal state
        If show_logs is not False and logs_interval is defined, it will print the logs every logs_interval seconds
        if show_logs is False it will not print the logs, will still pull the run state until it reaches terminal state
        """
        # TODO: rename sleep to pull_state_interval
        total_time = 0
        offset = 0
        last_pull_log_time = None
        logs_enabled = show_logs is not False
        state = self.state()
        if state not in mlrun.common.runtimes.constants.RunStates.terminal_states():
            logger.info(
                f"run {self.metadata.name} is not completed yet, waiting for it to complete",
                current_state=state,
            )
        while True:
            state = self.state()
            if (
                logs_enabled
                and logs_interval
                and state
                not in mlrun.common.runtimes.constants.RunStates.terminal_states()
                and (
                    last_pull_log_time is None
                    or (datetime.now() - last_pull_log_time).seconds > logs_interval
                )
            ):
                last_pull_log_time = datetime.now()
                state, offset = self.logs(watch=False, offset=offset)

            if state in mlrun.common.runtimes.constants.RunStates.terminal_states():
                if logs_enabled and logs_interval:
                    self.logs(watch=False, offset=offset)
                break
            time.sleep(sleep)
            total_time += sleep
            if timeout and total_time > timeout:
                raise mlrun.errors.MLRunTimeoutError(
                    "Run did not reach terminal state on time"
                )
        if logs_enabled and not logs_interval:
            self.logs(watch=False)
        if (
            raise_on_failure
            and state != mlrun.common.runtimes.constants.RunStates.completed
        ):
            raise mlrun.errors.MLRunRuntimeError(
                f"Task {self.metadata.name} did not complete (state={state})"
            )

        return state

    def abort(self):
        """abort the run"""
        db = mlrun.get_run_db()
        db.abort_run(self.metadata.uid, self.metadata.project)

    @staticmethod
    def create_uri(project: str, uid: str, iteration: Union[int, str], tag: str = ""):
        if tag:
            tag = f":{tag}"
        iteration = str(iteration)
        return f"{project}@{uid}#{iteration}{tag}"

    @staticmethod
    def parse_uri(uri: str) -> tuple[str, str, str, str]:
        """Parse the run's uri

        :param uri: run uri in the format of <project>@<uid>#<iteration>[:tag]
        :return: project, uid, iteration, tag
        """
        uri_pattern = mlrun.utils.regex.run_uri_pattern
        match = re.match(uri_pattern, uri)
        if not match:
            raise ValueError(
                "Uri not in supported format <project>@<uid>#<iteration>[:tag]"
            )
        group_dict = match.groupdict()
        return (
            group_dict["project"],
            group_dict["uid"],
            group_dict["iteration"],
            group_dict["tag"],
        )


class EntrypointParam(ModelObj):
    def __init__(
        self,
        name="",
        type=None,
        default=None,
        doc="",
        required=None,
        choices: Optional[list] = None,
    ):
        self.name = name
        self.type = type
        self.default = default
        self.doc = doc
        self.required = required
        self.choices = choices


class FunctionEntrypoint(ModelObj):
    def __init__(
        self,
        name="",
        doc="",
        parameters=None,
        outputs=None,
        lineno=-1,
        has_varargs=None,
        has_kwargs=None,
    ):
        self.name = name
        self.doc = doc
        self.parameters = [] if parameters is None else parameters
        self.outputs = [] if outputs is None else outputs
        self.lineno = lineno
        self.has_varargs = has_varargs
        self.has_kwargs = has_kwargs


def new_task(
    name=None,
    project=None,
    handler=None,
    params=None,
    hyper_params=None,
    param_file=None,
    selector=None,
    hyper_param_options=None,
    inputs=None,
    outputs=None,
    in_path=None,
    out_path=None,
    artifact_path=None,
    secrets=None,
    base=None,
    returns=None,
    retry=None,
) -> RunTemplate:
    """Creates a new task

    :param name:            task name
    :param project:         task project
    :param handler:         code entry-point/handler name
    :param params:          input parameters (dict)
    :param hyper_params:    dictionary of hyper parameters and list values, each
                            hyper param holds a list of values, the run will be
                            executed for every parameter combination (GridSearch)
    :param param_file:      a csv file with parameter combinations, first row hold
                            the parameter names, following rows hold param values
    :param selector:        selection criteria for hyper params e.g. "max.accuracy"
    :param hyper_param_options:   hyper parameter options, see: :py:class:`HyperParamOptions`
    :param inputs:          dictionary of input objects + optional paths (if path is
                            omitted the path will be the in_path/key)
    :param outputs:         dictionary of input objects + optional paths (if path is
                            omitted the path will be the out_path/key)
    :param in_path:         default input path/url (prefix) for inputs
    :param out_path:        default output path/url (prefix) for artifacts
    :param artifact_path:   default artifact output path
    :param secrets:         extra secrets specs, will be injected into the runtime
                            e.g. ['file=<filename>', 'env=ENV_KEY1,ENV_KEY2']
    :param base:            task instance to use as a base instead of a fresh new task instance
    :param returns:         List of log hints - configurations for how to log the returning values from the handler's
                            run (as artifacts or results). The list's length must be equal to the amount of returning
                            objects. A log hint may be given as:

                            * A string of the key to use to log the returning value as result or as an artifact. To
                              specify The artifact type, it is possible to pass a string in the following structure:
                              "<key> : <type>". Available artifact types can be seen in `mlrun.ArtifactType`. If no
                              artifact type is specified, the object's default artifact type will be used.
                            * A dictionary of configurations to use when logging. Further info per object type and
                              artifact type can be given there. The artifact key must appear in the dictionary as
                              "key": "the_key".
    :param retry:           Retry configuration for the run, can be a dict or an instance of mlrun.model.Retry.
    """

    if base:
        run = deepcopy(base)
    else:
        run = RunTemplate()
    run.metadata.name = name or run.metadata.name
    run.metadata.project = project or run.metadata.project
    run.spec.handler = handler or run.spec.handler
    run.spec.parameters = params or run.spec.parameters
    run.spec.inputs = inputs or run.spec.inputs
    run.spec.returns = returns or run.spec.returns
    run.spec.outputs = outputs or run.spec.outputs or []
    run.spec.input_path = in_path or run.spec.input_path
    run.spec.output_path = artifact_path or out_path or run.spec.output_path
    run.spec.secret_sources = secrets or run.spec.secret_sources or []

    run.spec.hyperparams = hyper_params or run.spec.hyperparams
    run.spec.hyper_param_options = hyper_param_options or run.spec.hyper_param_options
    run.spec.hyper_param_options.param_file = (
        param_file or run.spec.hyper_param_options.param_file
    )
    run.spec.hyper_param_options.selector = (
        selector or run.spec.hyper_param_options.selector
    )
    run.spec.retry = retry or run.spec.retry
    return run


class TargetPathObject:
    """Class configuring the target path
    This class will take consideration of a few parameters to create the correct end result path:

    - :run_id: if run_id is provided target will be considered as run_id mode which require to
        contain a {run_id} place holder in the path.
    - :is_single_file: if true then run_id must be the directory containing the output file
        or generated before the file name (run_id/output.file).
    - :base_path: if contains the place holder for run_id, run_id must not be None.
        if run_id passed and place holder doesn't exist the place holder will
        be generated in the correct place.
    """

    def __init__(
        self,
        base_path=None,
        run_id=None,
        is_single_file=False,
    ):
        self.run_id = run_id
        self.full_path_template = base_path
        if run_id is not None:
            if RUN_ID_PLACE_HOLDER not in self.full_path_template:
                if not is_single_file:
                    if self.full_path_template[-1] != "/":
                        self.full_path_template = self.full_path_template + "/"
                    self.full_path_template = (
                        self.full_path_template + RUN_ID_PLACE_HOLDER + "/"
                    )
                else:
                    dir_name_end = len(self.full_path_template)
                    if self.full_path_template[-1] != "/":
                        dir_name_end = self.full_path_template.rfind("/") + 1
                    updated_path = (
                        self.full_path_template[:dir_name_end]
                        + RUN_ID_PLACE_HOLDER
                        + "/"
                        + self.full_path_template[dir_name_end:]
                    )
                    self.full_path_template = updated_path
            else:
                if self.full_path_template[-1] != "/":
                    if self.full_path_template.endswith(RUN_ID_PLACE_HOLDER):
                        self.full_path_template = self.full_path_template + "/"
        else:
            if RUN_ID_PLACE_HOLDER in self.full_path_template:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Error when trying to create TargetPathObject with place holder '{run_id}' but no value."
                )

    def get_templated_path(self):
        return self.full_path_template

    def get_absolute_path(self, project_name=None):
        path = template_artifact_path(
            artifact_path=self.full_path_template,
            project=project_name,
        )
        return path.format(run_id=self.run_id) if self.run_id else path


class DataSource(ModelObj):
    """online or offline data source spec"""

    _dict_fields = [
        "kind",
        "name",
        "path",
        "attributes",
        "key_field",
        "time_field",
        "schedule",
        "online",
        "workers",
        "max_age",
        "start_time",
        "end_time",
    ]
    kind = None

    _fields_to_serialize = ["start_time", "end_time"]

    def __init__(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        attributes: Optional[dict[str, object]] = None,
        key_field: Optional[str] = None,
        time_field: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time: Optional[Union[datetime, str]] = None,
        end_time: Optional[Union[datetime, str]] = None,
    ):
        self.name = name
        self.path = str(path) if path is not None else None
        self.attributes = attributes or {}
        self.schedule = schedule
        self.key_field = key_field
        self.time_field = time_field
        self.start_time = start_time
        self.end_time = end_time

        self.online = None
        self.max_age = None
        self.workers = None
        self._secrets = {}

    def set_secrets(self, secrets):
        self._secrets = secrets

    def _serialize_field(
        self, struct: dict, field_name: Optional[str] = None, strip: bool = False
    ) -> typing.Any:
        value = super()._serialize_field(struct, field_name, strip)
        # We pull the field from self and not from struct because it was excluded from the struct when looping over
        # the fields to save.
        if field_name in ("start_time", "end_time") and isinstance(value, datetime):
            return value.isoformat()
        return value


class DataTargetBase(ModelObj):
    """data target spec, specify a destination for the feature set data"""

    _dict_fields = [
        "name",
        "kind",
        "path",
        "after_step",
        "attributes",
        "partitioned",
        "key_bucketing_number",
        "partition_cols",
        "time_partitioning_granularity",
        "max_events",
        "flush_after_seconds",
        "storage_options",
        "run_id",
        "schema",
    ]

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: Optional[dict] = None
    ):
        return super().from_dict(struct, fields=fields)

    def get_path(self):
        # polymorphism won't work here, because from_dict always returns an instance of the base type (DataTargetBase)
        if self.kind in ["stream", "kafka"]:
            return TargetPathObject(self.path or "")

        if self.path:
            is_single_file = hasattr(self, "is_single_file") and self.is_single_file()
            return TargetPathObject(self.path, self.run_id, is_single_file)
        else:
            return None

    def __init__(
        self,
        kind: Optional[str] = None,
        name: str = "",
        path=None,
        attributes: Optional[dict[str, str]] = None,
        after_step=None,
        partitioned: bool = False,
        key_bucketing_number: Optional[int] = None,
        partition_cols: Optional[list[str]] = None,
        time_partitioning_granularity: Optional[str] = None,
        max_events: Optional[int] = None,
        flush_after_seconds: Optional[int] = None,
        storage_options: Optional[dict[str, str]] = None,
        schema: Optional[dict[str, Any]] = None,
    ):
        self.name = name
        self.kind: str = kind
        self.path = path
        self.after_step = after_step
        self.attributes = attributes or {}
        self.last_written = None
        self.partitioned = partitioned
        self.key_bucketing_number = key_bucketing_number
        self.partition_cols = partition_cols
        self.time_partitioning_granularity = time_partitioning_granularity
        self.max_events = max_events
        self.flush_after_seconds = flush_after_seconds
        self.storage_options = storage_options
        self.run_id = None
        self.schema = schema


class FeatureSetProducer(ModelObj):
    """information about the task/job which produced the feature set data"""

    def __init__(self, kind=None, name=None, uri=None, owner=None, sources=None):
        self.kind = kind
        self.name = name
        self.owner = owner
        self.uri = uri
        self.sources = sources or {}


class DataTarget(DataTargetBase):
    """data target with extra status information (used in the feature-set/vector status)"""

    _dict_fields = [
        "name",
        "kind",
        "path",
        "attributes",
        "start_time",
        "online",
        "status",
        "updated",
        "size",
        "last_written",
        "run_id",
        "partitioned",
        "key_bucketing_number",
        "partition_cols",
        "time_partitioning_granularity",
    ]

    def __init__(
        self,
        kind: Optional[str] = None,
        name: str = "",
        path=None,
        online=None,
    ):
        super().__init__(kind, name, path)
        self.status = ""
        self.updated = None
        self.size = None
        self.online = online
        self.max_age = None
        self.start_time = None
        self.last_written = None
        self._producer = None
        self.producer = {}
        self.attributes = {}

    @property
    def producer(self) -> FeatureSetProducer:
        return self._producer

    @producer.setter
    def producer(self, producer):
        self._producer = self._verify_dict(producer, "producer", FeatureSetProducer)


class VersionedObjMetadata(ModelObj):
    def __init__(
        self,
        name: Optional[str] = None,
        tag: Optional[str] = None,
        uid: Optional[str] = None,
        project: Optional[str] = None,
        labels: Optional[dict[str, str]] = None,
        annotations: Optional[dict[str, str]] = None,
        updated=None,
    ):
        self.name = name
        self.tag = tag
        self.uid = uid
        self.project = project
        self.labels = labels or {}
        self.annotations = annotations or {}
        self.updated = updated



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for handling and sending notifications.
It is used by the SDK to send pipeline notifications, and by the cmdline to send publish git comments.
It is also used by the API to send run notifications via the run monitor.
For this reason, the module is in the general utils package, so it can be used by the SDK, CMD and the API.
"""

from .notification import *  # noqa
from .notification_pusher import *  # noqa



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import time

import mlrun.errors


def create_linear_backoff(base=2, coefficient=2, stop_value=120):
    """
    Create a generator of linear backoff. Check out usage example in test_helpers.py
    """
    x = 0
    comparison = min if coefficient >= 0 else max

    while True:
        next_value = comparison(base + x * coefficient, stop_value)
        yield next_value
        x += 1


def create_step_backoff(steps=None):
    """
    Create a generator of steps backoff.
    Example: steps = [[2, 5], [20, 10], [120, None]] will produce a generator in which the first 5
    values will be 2, the next 10 values will be 20 and the rest will be 120.
    :param steps: a list of lists [step_value, number_of_iteration_in_this_step]
    """
    steps = steps if steps is not None else [[2, 10], [10, 10], [120, None]]
    steps = iter(steps)

    # Get first step
    step = next(steps)
    while True:
        current_step_value, current_step_remain = step
        if current_step_remain == 0:
            # No more in this step, moving on
            step = next(steps)
        elif current_step_remain is None:
            # We are in the last step, staying here forever
            yield current_step_value
        elif current_step_remain > 0:
            # Still more remains in this step, just reduce the remaining number
            step[1] -= 1
            yield current_step_value


def create_exponential_backoff(base=2, max_value=120, scale_factor=1):
    """
    Create a generator of exponential backoff. Check out usage example in test_helpers.py
    :param base: exponent base
    :param max_value: max limit on the result
    :param scale_factor: factor to be used as linear scaling coefficient
    """
    exponent = 1
    while True:
        # This "complex" implementation (unlike the one in linear backoff) is to avoid exponent growing too fast and
        # risking going behind max_int
        next_value = scale_factor * (base**exponent)
        if next_value < max_value:
            exponent += 1
            yield next_value
        else:
            yield max_value


class Retryer:
    def __init__(
        self,
        backoff,
        timeout,
        logger,
        verbose,
        function,
        *args,
        fatal_exceptions=(),
        **kwargs,
    ):
        """
        Initialize function retryer with given *args and **kwargs.
        Tries to run it until success or timeout reached (timeout is optional)
        :param backoff: can either be a:
                - number (int / float) that will be used as interval.
                - generator of waiting intervals. (support next())
        :param timeout: pass None if timeout is not wanted, number of seconds if it is
        :param logger: a logger so we can log the failures
        :param verbose: whether to log the failure on each retry
        :param _function: function to run
        :param args: functions args
        :param fatal_exceptions: exception types that should not be retried
        :param kwargs: functions kwargs
        """
        self.backoff = backoff
        self.timeout = timeout
        self.logger = logger
        self.verbose = verbose
        self.function = function
        self.fatal_exceptions = tuple(fatal_exceptions or ())
        self.args = args
        self.kwargs = kwargs
        self.start_time = None
        self.last_exception = None
        self.first_interval = None

    def run(self):
        self._prepare()
        while not self._timeout_exceeded():
            next_interval = self.first_interval or next(self.backoff)
            result, exc, retry = self._perform_call(next_interval)

            if retry and type(exc) not in self.fatal_exceptions:
                time.sleep(next_interval)
            elif not exc:
                return result
            else:
                break

        self._raise_last_exception()

    def _prepare(self):
        self.start_time = time.monotonic()
        self.last_exception = None

        # Check if backoff is just a simple interval
        if isinstance(self.backoff, int) or isinstance(self.backoff, float):
            self.backoff = create_linear_backoff(base=self.backoff, coefficient=0)

        self.first_interval = next(self.backoff)
        if self.timeout and self.timeout <= self.first_interval:
            self.logger.warning(
                f"Timeout ({self.timeout}) must be higher than backoff ({self.first_interval})."
                f" Set timeout to be higher than backoff."
            )

    def _perform_call(self, next_interval):
        try:
            result = self.function(*self.args, **self.kwargs)
            return result, None, False
        except mlrun.errors.MLRunFatalFailureError as exc:
            raise exc.original_exception
        except Exception as exc:
            self.last_exception = exc
            return (
                None,
                self.last_exception,
                self._assert_failure_timeout(next_interval, exc),
            )

    def _assert_failure_timeout(self, next_interval, exc):
        self.last_exception = exc

        # If next interval is within allowed time period - wait on interval, abort otherwise
        if not self._timeout_exceeded(next_interval):
            if self.logger is not None and self.verbose:
                self.logger.debug(
                    f"Operation not yet successful, Retrying in {next_interval} seconds."
                    f" exc: {mlrun.errors.err_to_str(exc)}"
                )
            return True
        else:
            return False

    def _raise_last_exception(self):
        if self.logger is not None:
            self.logger.warning(
                f"Operation did not complete on time. last exception: {self.last_exception}"
            )

        raise mlrun.errors.MLRunRetryExhaustedError(
            f"Failed to execute command by the given deadline."
            f" last_exception: {self.last_exception},"
            f" function_name: {self.function.__name__},"
            f" timeout: {self.timeout}"
        ) from self.last_exception

    def _timeout_exceeded(self, next_interval=None):
        now = time.monotonic()
        if next_interval:
            now = now + next_interval
        return self.timeout is not None and now >= self.start_time + self.timeout


class AsyncRetryer(Retryer):
    async def run(self):
        self._prepare()
        while not self._timeout_exceeded():
            next_interval = self.first_interval or next(self.backoff)
            result, exc, retry = await self._perform_call(next_interval)
            if retry:
                await asyncio.sleep(next_interval)
            elif not exc:
                return result
            else:
                break

        self._raise_last_exception()

    async def _perform_call(self, next_interval):
        try:
            result = await self.function(*self.args, **self.kwargs)
            return result, None, False
        except mlrun.errors.MLRunFatalFailureError as exc:
            raise exc.original_exception
        except Exception as exc:
            return (
                None,
                self.last_exception,
                self._assert_failure_timeout(next_interval, exc),
            )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib

from pyspark.sql.functions import udf
from pyspark.sql.types import StringType


def _hash_list(*list_to_hash):
    list_to_hash = [str(element) for element in list_to_hash]
    str_concatted = "".join(list_to_hash)
    sha1 = hashlib.sha1()
    sha1.update(str_concatted.encode("utf8"))
    return sha1.hexdigest()


def _redis_stringify_key(*args):
    if len(args) == 1:
        key_list = args[0]
    else:
        key_list = list(args)
    suffix = "}:static"
    if isinstance(key_list, list):
        if len(key_list) >= 3:
            return str(key_list[0]) + "." + _hash_list(*key_list[1:]) + suffix
        if len(key_list) == 2:
            return str(key_list[0]) + "." + str(key_list[1]) + suffix
        return str(key_list[0]) + suffix
    return str(key_list) + suffix


hash_and_concat_v3io_udf = udf(_hash_list, StringType())
hash_and_concat_redis_udf = udf(_redis_stringify_key, StringType())



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util as imputil
import inspect
import io
import json
import os
import socket
import sys
import tempfile
import threading
import traceback
from contextlib import redirect_stdout
from copy import copy
from io import StringIO
from os import environ, remove
from pathlib import Path
from subprocess import PIPE, Popen
from sys import executable
from typing import Optional

from nuclio import Event

import mlrun
import mlrun.common.constants as mlrun_constants
from mlrun.lists import RunList

from ..errors import err_to_str
from ..execution import MLClientCtx
from ..model import RunObject
from ..utils import get_handler_extended, get_in, logger, set_paths
from ..utils.clones import extract_source
from .base import BaseRuntime
from .kubejob import KubejobRuntime
from .remotesparkjob import RemoteSparkRuntime
from .utils import RunError, global_context, log_std


class ParallelRunner:
    def _get_trackers_manager(self):
        """
        useful to import and call get_trackers_manager from mlrun.track in order to avoid circular imports
        or imports in multiple spots mid-code
        :return: trackers_manager
        """
        from mlrun.track import TrackerManager

        return TrackerManager()

    def _get_handler(
        self, handler: str, context: MLClientCtx, embed_in_sys: bool = True
    ):
        return handler

    def _get_dask_client(self, options):
        from distributed import Client

        if options.dask_cluster_uri:
            function = mlrun.import_function(options.dask_cluster_uri)
            return function.client, function.metadata.name
        return Client(), None

    def _parallel_run_many(
        self, generator, execution: MLClientCtx, runobj: RunObject
    ) -> RunList:
        # TODO: this flow assumes we use dask - move it to dask runtime
        from distributed import as_completed

        if self.spec.build.source and generator.options.dask_cluster_uri:
            # the attached dask cluster will not have the source code when we clone the git on run
            raise mlrun.errors.MLRunRuntimeError(
                "Cannot load source code into remote Dask at runtime use, "
                "function.deploy() to add the code into the image instead"
            )
        results = RunList()
        tasks = generator.generate(runobj)
        handler = runobj.spec.handler
        self._force_handler(handler)
        set_paths(self.spec.pythonpath)
        handler = self._get_handler(handler, execution, embed_in_sys=False)

        client, function_name = self._get_dask_client(generator.options)
        parallel_runs = generator.options.parallel_runs or 4
        queued_runs = 0
        num_errors = 0

        def process_result(future):
            nonlocal num_errors
            resp, sout, serr = future.result()
            runobj = RunObject.from_dict(resp)
            try:
                log_std(self._db_conn, runobj, sout, serr, skip=self.is_child)
                resp = self._update_run_state(resp)
            except RunError as err:
                resp = self._update_run_state(resp, err=err_to_str(err))
                num_errors += 1
            results.append(resp)
            if num_errors > generator.max_errors:
                logger.error("Max errors reached, stopping iterations!")
                return True
            run_results = resp["status"].get("results", {})
            stop = generator.eval_stop_condition(run_results)
            if stop:
                logger.info(
                    f"Reached early stop condition ({generator.options.stop_condition}), stopping iterations!"
                )
            return stop

        completed_iter = as_completed([])
        for task in tasks:
            task_struct = task.to_dict()
            project = get_in(task_struct, "metadata.project")
            uid = get_in(task_struct, "metadata.uid")
            iter = get_in(task_struct, "metadata.iteration", 0)
            mlrun.get_run_db().store_run(
                task_struct, uid=uid, project=project, iter=iter
            )
            resp = client.submit(
                remote_handler_wrapper, task.to_json(), handler, self.spec.workdir
            )
            completed_iter.add(resp)
            queued_runs += 1
            if queued_runs >= parallel_runs:
                future = next(completed_iter)
                early_stop = process_result(future)
                queued_runs -= 1
                if early_stop:
                    break

        for future in completed_iter:
            process_result(future)

        client.close()
        if function_name and generator.options.teardown_dask:
            logger.info("Tearing down the dask cluster..")
            mlrun.get_run_db().delete_runtime_resources(
                project=self.metadata.project,
                kind=mlrun.runtimes.RuntimeKinds.dask,
                object_id=function_name,
                force=True,
            )

        return results


def remote_handler_wrapper(task, handler, workdir=None):
    if task and not isinstance(task, dict):
        task = json.loads(task)

    context = MLClientCtx.from_dict(
        task,
        autocommit=False,
        host=socket.gethostname(),
    )
    runobj = RunObject.from_dict(task)

    sout, serr = exec_from_params(handler, runobj, context, workdir)
    return context.to_dict(), sout, serr


class HandlerRuntime(BaseRuntime, ParallelRunner):
    kind = "handler"

    def _run(self, runobj: RunObject, execution: MLClientCtx):
        handler = runobj.spec.handler
        self._force_handler(handler)
        tmp = tempfile.NamedTemporaryFile(suffix=".json", delete=False).name
        environ["MLRUN_META_TMPFILE"] = tmp
        set_paths(self.spec.pythonpath)

        context = MLClientCtx.from_dict(
            runobj.to_dict(),
            rundb=self.spec.rundb,
            autocommit=False,
            tmp=tmp,
            host=socket.gethostname(),
        )
        global_context.set(context)
        # Running tracking services pre run to detect if some of them should be used:
        trackers_manager = self._get_trackers_manager()
        trackers_manager.pre_run(context)
        sout, serr = exec_from_params(handler, runobj, context, self.spec.workdir)
        context = trackers_manager.post_run(context)
        log_std(self._db_conn, runobj, sout, serr, show=False)
        return context.to_dict()


class LocalRuntime(BaseRuntime, ParallelRunner):
    kind = "local"
    _is_remote = False

    def to_job(self, image="", func_name: Optional[str] = None):
        struct = self.to_dict()
        obj = KubejobRuntime.from_dict(struct)
        obj.kind = "job"  # Ensure kind is set to 'job' for KubejobRuntime
        if func_name:
            obj.metadata.name = func_name
        if image:
            obj.spec.image = image
        return obj

    def with_source_archive(self, source, workdir=None, handler=None, target_dir=None):
        """load the code from git/tar/zip archive at runtime or build

        :param source:      valid path to git, zip, or tar file, e.g.
                            git://github.com/mlrun/something.git
                            http://some/url/file.zip
        :param handler:     default function handler
        :param workdir:     working dir relative to the archive root (e.g. './subdir') or absolute
        :param target_dir:  local target dir for repo clone (by default its <current-dir>/code)
        """
        self.spec.build.source = source
        self.spec.build.load_source_on_run = True
        if handler:
            self.spec.default_handler = handler
        if workdir:
            self.spec.workdir = workdir
        if target_dir:
            self.spec.build.source_code_target_dir = target_dir

    def is_deployed(self):
        return True

    def _get_handler(
        self, handler: str, context: MLClientCtx, embed_in_sys: bool = True
    ):
        command = self.spec.command
        if not command and self.spec.build.functionSourceCode:
            # if the code is embedded in the function object extract or find it
            command, _ = mlrun.run.load_func_code(self)
        return load_module(command, handler, context, embed_in_sys=embed_in_sys)

    def _pre_run(self, runobj: RunObject, execution: MLClientCtx):
        workdir = self.spec.workdir
        execution._current_workdir = workdir
        execution._old_workdir = None

        # _is_run_local is set when the user specifies local=True in run()
        # in this case we don't want to extract the source code and contaminate the user's local dir
        if self.spec.build.source and not hasattr(self, "_is_run_local"):
            target_dir = extract_source(
                self.spec.build.source,
                self.spec.build.source_code_target_dir,
                secrets=execution._secrets_manager,
            )
            if workdir and not workdir.startswith("/"):
                execution._current_workdir = os.path.join(target_dir, workdir)
            else:
                execution._current_workdir = workdir or target_dir

        if execution._current_workdir:
            execution._old_workdir = os.getcwd()
            workdir = os.path.realpath(execution._current_workdir)
            set_paths(workdir)
            os.chdir(workdir)
        else:
            set_paths(os.path.realpath("."))

        if (
            runobj.metadata.labels.get(mlrun_constants.MLRunInternalLabels.kind)
            == RemoteSparkRuntime.kind
            and environ["MLRUN_SPARK_CLIENT_IGZ_SPARK"] == "true"
        ):
            from mlrun.runtimes.remotesparkjob import igz_spark_pre_hook

            igz_spark_pre_hook()

    def _post_run(self, results, execution: MLClientCtx):
        if execution._old_workdir:
            os.chdir(execution._old_workdir)

    def _run(self, runobj: RunObject, execution: MLClientCtx):
        # we define a tmp file for mlrun to log its run, for easy access later
        environ["MLRUN_EXEC_CONFIG"] = runobj.to_json()
        tmp = tempfile.NamedTemporaryFile(suffix=".json", delete=False).name
        environ["MLRUN_META_TMPFILE"] = tmp
        if self.spec.rundb:
            environ["MLRUN_DBPATH"] = self.spec.rundb

        handler = runobj.spec.handler
        handler_str = handler or "main"
        logger.debug(f"Starting local run: {self.spec.command} # {handler_str}")
        pythonpath = self.spec.pythonpath

        if handler:
            set_paths(pythonpath)

            context = MLClientCtx.from_dict(
                runobj.to_dict(),
                rundb=self.spec.rundb,
                autocommit=False,
                tmp=tmp,
                host=socket.gethostname(),
            )
            try:
                fn = self._get_handler(handler, context)
                global_context.set(context)
                # Running tracking services pre run to detect if some of them should be used:
                trackers_manager = self._get_trackers_manager()
                trackers_manager.pre_run(context)
                sout, serr = exec_from_params(fn, runobj, context)
                # If trackers where used, this is where we log all data collected to MLRun
                context = trackers_manager.post_run(context)
                log_std(
                    self._db_conn, runobj, sout, serr, skip=self.is_child, show=False
                )
                return context.to_dict()

            # if RunError was raised it means that the error was raised as part of running the function
            except RunError as err:
                raise err
            except Exception as exc:
                logger.error(f"Run error, {traceback.format_exc()}")
                raise RunError(
                    "Failed on pre-loading / post-running of the function"
                ) from exc

        else:
            command = self.spec.command
            command = command.format(**runobj.spec.parameters)
            logger.info(f"Handler was not provided running main ({command})")
            arg_list = command.split()
            if self.spec.mode == "pass":
                cmd = arg_list
            else:
                cmd = [executable, "-u"] + arg_list

            env = None
            if pythonpath:
                if "PYTHONPATH" in environ:
                    pythonpath = f"{environ['PYTHONPATH']}:{pythonpath}"
                env = {"PYTHONPATH": pythonpath}
            if runobj.spec.verbose:
                if not env:
                    env = {}
                env["MLRUN_LOG_LEVEL"] = "DEBUG"

            args = self.spec.args
            if args:
                new_args = []
                for arg in args:
                    arg = arg.format(**runobj.spec.parameters)
                    new_args.append(arg)
                args = new_args
            # Running tracking services pre run to detect if some of them should be used:
            trackers_manager = self._get_trackers_manager()
            trackers_manager.pre_run(execution)
            sout, serr = run_exec(cmd, args, env=env, cwd=execution._current_workdir)

            run_obj_dict = runobj.to_dict()  # default value
            if os.path.isfile(tmp):
                with open(tmp) as fp:
                    resp = fp.read()
                remove(tmp)
                if resp:
                    run_obj_dict = json.loads(resp)
                else:
                    logger.debug("Empty context tmp file")
            else:
                logger.info("No context file found")

            # If trackers where used, this is where we log all data collected to MLRun
            run_obj_dict = trackers_manager.post_run(run_obj_dict)
            log_std(self._db_conn, runobj, sout, serr, skip=self.is_child, show=False)
            return run_obj_dict


def load_module(
    file_name: str,
    handler: str,
    context: MLClientCtx,
    embed_in_sys: bool = True,
):
    """
    Load module from filename
    :param file_name:       The module path to load
    :param handler:         The callable to load
    :param context:         Execution context
    :param embed_in_sys:    Embed the file-named module in sys.modules. This is not persistent with remote
                            environments and therefore can effect pickling.
    """
    module = None
    if file_name:
        path = Path(file_name)
        mod_name = path.name
        if path.suffix:
            mod_name = mod_name[: -len(path.suffix)]
        spec = imputil.spec_from_file_location(mod_name, file_name)
        if spec is None:
            raise RunError(f"Cannot import from {file_name!r}")
        module = imputil.module_from_spec(spec)
        if embed_in_sys:
            sys.modules[mod_name] = module
        spec.loader.exec_module(module)

    class_args = {}
    if context:
        class_args = copy(context._parameters.get("_init_args", {}))

    return get_handler_extended(
        handler,
        context,
        class_args,
        namespaces=module,
        reload_modules=context._reset_on_run,
    )


def run_exec(cmd, args, env=None, cwd=None):
    if args:
        cmd += args
    if env and "SYSTEMROOT" in os.environ:
        env["SYSTEMROOT"] = os.environ["SYSTEMROOT"]
    print("Running:", cmd)
    process = Popen(
        cmd, stdout=PIPE, stderr=PIPE, env=os.environ, cwd=cwd, universal_newlines=True
    )

    def read_stderr(stderr):
        while True:
            nextline = process.stderr.readline()
            if not nextline:
                break
            stderr.write(nextline)

    # ML-3710. We must read stderr in a separate thread to drain the stderr pipe so that the spawned process won't
    # hang if it tries to write more to stderr than the buffer size (default of approx 8kb).
    with io.StringIO() as stderr:
        stderr_consumer_thread = threading.Thread(target=read_stderr, args=[stderr])
        stderr_consumer_thread.start()

        with io.StringIO() as stdout:
            while True:
                nextline = process.stdout.readline()
                if not nextline:
                    break
                print(nextline, end="")
                sys.stdout.flush()
                stdout.write(nextline)
            out = stdout.getvalue()

        stderr_consumer_thread.join()
        err = stderr.getvalue()

    process.wait()

    # if we return anything for err, the caller will assume that the process failed
    code = process.poll()
    err = "" if code == 0 else err

    return out, err


class _DupStdout:
    def __init__(self):
        self.terminal = sys.stdout
        self.buf = StringIO()

    def write(self, message):
        self.terminal.write(message)
        self.buf.write(message)

    def flush(self):
        self.terminal.flush()


def exec_from_params(handler, runobj: RunObject, context: MLClientCtx, cwd=None):
    old_level = logger.level
    try:
        if runobj.spec.verbose:
            logger.set_logger_level("DEBUG")

        # Prepare the inputs type hints (user may pass type hints as part of the input keys):
        runobj.spec.extract_type_hints_from_inputs()
        # Read the keyword arguments to pass to the function (combining params and inputs from the run spec):
        kwargs = get_func_arg(handler, runobj, context)

        stdout = _DupStdout()
        err = ""
        val = None
        old_dir = os.getcwd()
        commit = True
        with redirect_stdout(stdout):
            context.set_logger_stream(stdout)
            try:
                if cwd:
                    os.chdir(cwd)
                # Apply the MLRun handler decorator for parsing inputs using type hints and logging outputs using
                # log hints (Expected behavior: inputs are being parsed when they have type hints in code or given
                # by user. Outputs are logged only if log hints are provided by the user):
                if mlrun.mlconf.packagers.enabled:
                    val = mlrun.handler(
                        inputs=(
                            runobj.spec.inputs_type_hints
                            if runobj.spec.inputs_type_hints
                            else True  # True will use type hints if provided in user's code.
                        ),
                        outputs=(
                            runobj.spec.returns
                            if runobj.spec.returns
                            else None  # None will turn off outputs logging.
                        ),
                    )(handler)(**kwargs)
                else:
                    val = handler(**kwargs)
                context.set_state("completed", commit=False)
            except mlrun.errors.MLRunTaskCancelledError as exc:
                logger.warning("Run was aborted", err=err_to_str(exc))
                # Run was aborted, the state run state is updated by the abort job, no need to commit again
                context.set_state(
                    mlrun.common.runtimes.constants.RunStates.aborted, commit=False
                )
                commit = False
            except Exception as exc:
                err = err_to_str(exc)
                logger.error(f"Execution error, {traceback.format_exc()}")
                context.set_state(error=err, commit=False)

        stdout.flush()
        if cwd:
            os.chdir(old_dir)
        context.set_logger_stream(sys.stdout)
        if val:
            context.log_result("return", val)

        if commit:
            # completion will be ignored if error is set
            context.commit(completed=True)

    finally:
        logger.set_logger_level(old_level)
    return stdout.buf.getvalue(), err


def get_func_arg(handler, runobj: RunObject, context: MLClientCtx, is_nuclio=False):
    params = runobj.spec.parameters or {}
    inputs = runobj.spec.inputs or {}
    kwargs = {}
    args = inspect.signature(handler).parameters

    def _get_input_value(input_key: str):
        input_obj = context.get_input(input_key, inputs[input_key])
        # If there is no type hint annotation but there is a default value and its type is string, point the data
        # item to local downloaded file path (`local()` returns the downloaded temp path string):
        if args[input_key].annotation is inspect.Parameter.empty and isinstance(
            args[input_key].default, str
        ):
            return input_obj.local()
        else:
            return input_obj

    for key in args.keys():
        if key == "context":
            kwargs[key] = context
        elif is_nuclio and key == "event":
            kwargs[key] = Event(runobj.to_dict())
        elif key in params:
            kwargs[key] = copy(params[key])
        elif key in inputs:
            kwargs[key] = _get_input_value(key)

    list_of_params = list(args.values())
    if len(list_of_params) == 0:
        return kwargs

    # get the last parameter, as **kwargs can only be last in the function's parameters list
    last_param = list_of_params[-1]
    # VAR_KEYWORD meaning : A dict of keyword arguments that aren’t bound to any other parameter.
    # This corresponds to a **kwargs parameter in a Python function definition.
    if last_param.kind == last_param.VAR_KEYWORD:
        # if handler has **kwargs, pass all parameters provided by the user to the handler which were not already set
        # as part of the previous loop which handled all parameters which were explicitly defined in the handler
        for key in params:
            if key not in kwargs:
                kwargs[key] = copy(params[key])
        for key in inputs:
            if key not in kwargs:
                kwargs[key] = _get_input_value(key)
    return kwargs



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from datetime import datetime, timedelta

import requests

import mlrun.errors
from mlrun.utils import logger


class TokenProvider(ABC):
    @abstractmethod
    def get_token(self):
        pass

    @abstractmethod
    def is_iguazio_session(self):
        pass


class StaticTokenProvider(TokenProvider):
    def __init__(self, token: str):
        self.token = token

    def get_token(self):
        return self.token

    def is_iguazio_session(self):
        return mlrun.platforms.iguazio.is_iguazio_session(self.token)


class OAuthClientIDTokenProvider(TokenProvider):
    def __init__(
        self, token_endpoint: str, client_id: str, client_secret: str, timeout=5
    ):
        if not token_endpoint or not client_id or not client_secret:
            raise mlrun.errors.MLRunValueError(
                "Invalid client_id configuration for authentication. Must provide token endpoint, client-id and secret"
            )
        self.token_endpoint = token_endpoint
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout = timeout

        # Since we're only issuing POST requests, which are actually a disguised GET, then it's ok to allow retries
        # on them.
        self._session = mlrun.utils.HTTPSessionWithRetry(
            retry_on_post=True,
            verbose=True,
        )

        self._cleanup()
        self._refresh_token_if_needed()

    def get_token(self):
        self._refresh_token_if_needed()
        return self.token

    def is_iguazio_session(self):
        return False

    def _cleanup(self):
        self.token = self.token_expiry_time = self.token_refresh_time = None

    def _refresh_token_if_needed(self):
        now = datetime.now()
        if self.token:
            if self.token_refresh_time and now <= self.token_refresh_time:
                return self.token

            # We only cleanup if token was really expired - even if we fail in refreshing the token, we can still
            # use the existing one given that it's not expired.
            if now >= self.token_expiry_time:
                self._cleanup()

        self._issue_token_request()
        return self.token

    def _issue_token_request(self, raise_on_error=False):
        try:
            headers = {"Content-Type": "application/x-www-form-urlencoded"}
            request_body = {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }
            response = self._session.request(
                "POST",
                self.token_endpoint,
                timeout=self.timeout,
                headers=headers,
                data=request_body,
            )
        except requests.RequestException as exc:
            error = f"Retrieving token failed: {mlrun.errors.err_to_str(exc)}"
            if raise_on_error:
                raise mlrun.errors.MLRunRuntimeError(error) from exc
            else:
                logger.warning(error)
                return

        if not response.ok:
            error = "No error available"
            if response.content:
                try:
                    data = response.json()
                    error = data.get("error")
                except Exception:
                    pass
            logger.warning(
                "Retrieving token failed", status=response.status_code, error=error
            )
            if raise_on_error:
                mlrun.errors.raise_for_status(response)
            return

        self._parse_response(response.json())

    def _parse_response(self, data: dict):
        # Response is described in https://datatracker.ietf.org/doc/html/rfc6749#section-4.4.3
        # According to spec, there isn't a refresh token - just the access token and its expiry time (in seconds).
        self.token = data.get("access_token")
        expires_in = data.get("expires_in")
        if not self.token or not expires_in:
            token_str = "****" if self.token else "missing"
            logger.warning(
                "Failed to parse token response", token=token_str, expires_in=expires_in
            )
            return

        now = datetime.now()
        self.token_expiry_time = now + timedelta(seconds=expires_in)
        self.token_refresh_time = now + timedelta(seconds=expires_in / 2)
        logger.info(
            "Successfully retrieved client-id token",
            expires_in=expires_in,
            expiry=str(self.token_expiry_time),
            refresh=str(self.token_refresh_time),
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from sklearn.metrics import roc_auc_score, roc_curve

from mlrun.artifacts import Artifact, PlotlyArtifact

from ..plan import MLPlanStages, MLPlotPlan
from ..utils import MLTypes, MLUtils


class ROCCurvePlan(MLPlotPlan):
    """
    Plan for producing a receiver operating characteristic (ROC) - a plot that shows the connection / trade-off between
    clinical sensitivity and specificity for every possible cut-off for a test or a combination of tests.
    """

    _ARTIFACT_NAME = "roc-curves"

    def __init__(
        self,
        pos_label: Optional[Union[str, int]] = None,
        sample_weight: np.ndarray = None,
        drop_intermediate: bool = True,
        average: str = "macro",
        max_fpr: Optional[float] = None,
        multi_class: str = "raise",
        labels: Optional[list[str]] = None,
    ):
        """
        Initialize a receiver operating characteristic plan with the given configuration.

        To read more about the parameters, head to the SciKit-Learn docs at:

        * https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_curve.html
        * https://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_auc_score.html

        :param pos_label:         The label of the positive class. When None, if 'y_true' (y) is in {-1, 1} or {0, 1},
                                  'pos_label' is set to 1, otherwise an error will be raised. Default: None.
        :param sample_weight:     Sample weights to apply.
        :param drop_intermediate: Whether to drop some suboptimal thresholds which would not appear on a plotted ROC
                                  curve. Default: True.
        :param average:           Determines the type of averaging performed on the data. If None, the scores for each
                                  class are returned. Default: "macro".
        :param max_fpr:           For multiclass it should be equal to 1 or None. If not None, the standardized partial
                                  AUC [2] over the range [0, max_fpr] is returned. Default: None.
        :param multi_class:       Only used for multiclass targets. Determines the type of configuration to use. Can be
                                  one of {'raise', 'ovr', 'ovo'}. Default: "raise".
        :param labels:            Only used for multiclass targets. List of labels that index the classes in 'y_pred'.
                                  If None, the labels found in 'y_true' (y) will be used. Default: None.
        """
        # Store the parameters:
        self._pos_label = pos_label
        self._sample_weight = sample_weight
        self._drop_intermediate = drop_intermediate
        self._average = average
        self._max_fpr = max_fpr
        self._multi_class = multi_class
        self._labels = labels

        # Continue the initialization for the MLPlan:
        super().__init__(need_probabilities=True)

    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether or not the plan is fit for production by the given stage and prediction probabilities. The
        roc curve is ready only post prediction probabilities.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        return stage == MLPlanStages.POST_PREDICT and is_probabilities

    def produce(
        self,
        y: MLTypes.DatasetType,
        y_pred: MLTypes.DatasetType = None,
        model: MLTypes.ModelType = None,
        x: MLTypes.DatasetType = None,
        **kwargs,
    ) -> dict[str, Artifact]:
        """
        Produce the roc curve according to the ground truth (y) and predictions (y_pred) values. If predictions are not
        available, the model and a dataset can be given to produce them.

        :param y:      The ground truth values.
        :param y_pred: The predictions values.
        :param model:  Model to produce the predictions.
        :param x:      Input dataset to produce the predictions.

        :return: The produced roc curve artifact in an artifacts dictionary.
        """
        # Calculate the predictions if needed:
        y_pred = self._calculate_predictions(y_pred=y_pred, model=model, x=x)

        # Convert to DataFrame:
        y = MLUtils.to_dataframe(dataset=y)
        y_pred = MLUtils.to_dataframe(dataset=y_pred)

        # One hot encode the labels in order to plot them
        y_one_hot = pd.get_dummies(y, columns=y.columns.to_list())

        # Create an empty figure:
        fig = go.Figure()
        fig.add_shape(type="line", line={"dash": "dash"}, x0=0, x1=1, y0=0, y1=1)

        # Iteratively add new lines every time we compute a new class:
        for i in range(y_pred.shape[1]):
            class_i_true = y_one_hot.iloc[:, i]
            class_i_pred = y_pred.iloc[:, i]
            fpr, tpr, _ = roc_curve(
                class_i_true,
                class_i_pred,
                pos_label=self._pos_label,
                sample_weight=self._sample_weight,
                drop_intermediate=self._drop_intermediate,
            )
            auc_score = roc_auc_score(
                class_i_true,
                class_i_pred,
                average=self._average,
                sample_weight=self._sample_weight,
                max_fpr=self._max_fpr,
                multi_class=self._multi_class,
                labels=self._labels,
            )
            name = f"{y_one_hot.columns[i]} (AUC={auc_score:.2f})"
            fig.add_trace(go.Scatter(x=fpr, y=tpr, name=name, mode="lines"))

        # Configure the layout:
        fig.update_layout(
            xaxis_title="False Positive Rate",
            yaxis_title="True Positive Rate",
            yaxis={"scaleanchor": "x", "scaleratio": 1},
            xaxis={"constrain": "domain"},
            width=700,
            height=500,
        )

        # Creating the plot artifact:
        self._artifacts[self._ARTIFACT_NAME] = PlotlyArtifact(
            key=self._ARTIFACT_NAME,
            figure=fig,
        )

        return self._artifacts



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import storey

import mlrun.errors


class ChoiceByField(storey.Choice):
    """
    Selects downstream outlets to route each event based on a predetermined field.
    :param field_name: event field name that contains the step name or names of the desired outlet or outlets
    """

    def __init__(self, field_name: Union[str, list[str]], **kwargs):
        self.field_name = field_name
        super().__init__(**kwargs)

    def select_outlets(self, event):
        # Case 1: Missing field
        if self.field_name not in event:
            raise mlrun.errors.MLRunRuntimeError(
                f"Field '{self.field_name}' is not contained in the event keys {list(event.keys())}."
            )

        outlet = event[self.field_name]

        # Case 2: Field exists but is None
        if outlet is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Field '{self.field_name}' exists but its value is None."
            )

        # Case 3: Invalid type
        if not isinstance(outlet, (str, list, tuple)):
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                f"Field '{self.field_name}' must be a string or list of strings "
                f"but is instead of type '{type(outlet).__name__}'."
            )

        outlets = [outlet] if isinstance(outlet, str) else outlet

        # Case 4: Empty list or tuple
        if not outlets:
            raise mlrun.errors.MLRunRuntimeError(
                f"The value of the key '{self.field_name}' cannot be an empty {type(outlets).__name__}."
            )

        return outlets



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import typing
from base64 import b64decode

import yaml
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import OperationFailed
from databricks.sdk.service.compute import ClusterSpec
from databricks.sdk.service.jobs import (
    Run,
    RunResultState,
    RunTask,
    SparkPythonTask,
    SubmitTask,
)

import mlrun
from mlrun.errors import MLRunRuntimeError, MLRunTaskCancelledError

credentials_path = "/mlrun/databricks_credentials.yaml"


def get_task(databricks_run: Run) -> RunTask:
    if len(databricks_run.tasks) == 0:
        raise MLRunRuntimeError(
            "Cannot find tasks related to the Databricks job run ID in Databricks environment."
        )
    elif len(databricks_run.tasks) > 1:
        raise MLRunRuntimeError(
            "More than one task related to the Databricks job run ID has been found."
            " Did you manually run any tasks from this job?"
        )
    return databricks_run.tasks[0]


def log_artifacts_by_dbfs_json(
    context: mlrun.MLClientCtx,
    workspace: WorkspaceClient,
    artifact_json_path: str,
    databricks_run_name: str,
):
    if not workspace.dbfs.exists(artifact_json_path):
        return
    context.logger.info(f"Artifacts found. Run name: {databricks_run_name}")
    with workspace.dbfs.open(artifact_json_path, read=True) as artifact_file:
        artifact_json = json.load(artifact_file)
    for artifact_name, artifact_path in artifact_json.items():
        fixed_artifact_path = artifact_path
        if artifact_path.startswith("/dbfs"):
            fixed_artifact_path = artifact_path.replace("/dbfs", "dbfs://", 1)
        # for pyspark format:
        elif artifact_path.startswith("dbfs:/") and not artifact_path.startswith(
            "dbfs://"
        ):
            fixed_artifact_path = artifact_path.replace("dbfs:/", "dbfs:///", 1)
        elif not artifact_path.startswith("dbfs:///"):
            context.logger.error(
                f"Can not log artifact: {artifact_name}: {artifact_path}"
            )
            continue
        context.log_artifact(
            artifact_name, local_path=fixed_artifact_path, upload=False
        )


def save_credentials(
    workspace: WorkspaceClient,
    waiter,
    host: str,
    token: str,
    cluster_id: typing.Optional[str],
    is_finished: bool,
):
    databricks_run = workspace.jobs.get_run(run_id=waiter.run_id)
    task_run_id = get_task(databricks_run=databricks_run).run_id
    credentials = {
        "DATABRICKS_HOST": host,
        "DATABRICKS_TOKEN": token,
        "TASK_RUN_ID": task_run_id,
        "IS_FINISHED": is_finished,
    }
    if cluster_id:
        credentials["DATABRICKS_CLUSTER_ID"] = cluster_id

    with open(credentials_path, "w") as yaml_file:
        yaml.safe_dump(credentials, yaml_file, default_flow_style=False)


def run_mlrun_databricks_job(
    context: mlrun.MLClientCtx,
    task_parameters: dict,
    **kwargs,
):
    spark_app_code = task_parameters["spark_app_code"]
    token_key = task_parameters.get("token_key", "DATABRICKS_TOKEN")
    databricks_token = mlrun.get_secret_or_env(key=token_key)
    host = mlrun.get_secret_or_env(key="DATABRICKS_HOST")
    timeout_minutes = task_parameters.get("timeout_minutes", 20)
    number_of_workers = task_parameters.get("number_of_workers", 1)
    new_cluster_spec = task_parameters.get("new_cluster_spec")
    artifact_json_path = task_parameters.get("artifact_json_path")
    current_time = datetime.datetime.utcnow()
    run_time = current_time.strftime("%H_%M_%S_%f")
    databricks_run_name = task_parameters.get("databricks_run_name", "mlrun_task_")
    databricks_run_name = f"{databricks_run_name}_{run_time}"
    logger = context.logger
    workspace = WorkspaceClient(token=databricks_token)
    script_path_on_dbfs = (
        f"/home/{workspace.current_user.me().user_name}/mlrun_databricks_runtime/"
        f"{databricks_run_name}.py"
    )
    spark_app_code = b64decode(spark_app_code)
    if workspace.dbfs.exists(artifact_json_path):
        workspace.dbfs.delete(artifact_json_path)
    with workspace.dbfs.open(script_path_on_dbfs, write=True, overwrite=True) as f:
        f.write(spark_app_code)

    def print_status(run: Run):
        statuses = [f"{t.task_key}: {t.state.life_cycle_state}" for t in run.tasks]
        logger.info(f'Workflow intermediate status: {", ".join(statuses)}')

    try:
        cluster_id = mlrun.get_secret_or_env("DATABRICKS_CLUSTER_ID")
        submit_task_kwargs = {}
        if cluster_id:
            logger.info("Running with an existing cluster", cluster_id=cluster_id)
            submit_task_kwargs["existing_cluster_id"] = cluster_id
        else:
            logger.info("Running with a new cluster")
            cluster_spec_kwargs = {
                "spark_version": workspace.clusters.select_spark_version(
                    long_term_support=True
                ),
                "node_type_id": workspace.clusters.select_node_type(local_disk=True),
                "num_workers": number_of_workers,
            }
            if new_cluster_spec:
                cluster_spec_kwargs.update(new_cluster_spec)
            submit_task_kwargs["new_cluster"] = ClusterSpec(**cluster_spec_kwargs)
        waiter = workspace.jobs.submit(
            run_name=databricks_run_name,
            tasks=[
                SubmitTask(
                    task_key=databricks_run_name,
                    spark_python_task=SparkPythonTask(
                        python_file=f"dbfs:{script_path_on_dbfs}",
                        parameters=[json.dumps(kwargs)],
                    ),
                    **submit_task_kwargs,
                )
            ],
        )
        logger.info(f"Starting to poll: {waiter.run_id}")
        save_credentials(
            workspace=workspace,
            waiter=waiter,
            host=host,
            token=databricks_token,
            cluster_id=cluster_id,
            is_finished=False,
        )
        try:
            run = waiter.result(
                timeout=datetime.timedelta(minutes=timeout_minutes),
                callback=print_status,
            )
            log_artifacts_by_dbfs_json(
                context=context,
                workspace=workspace,
                artifact_json_path=artifact_json_path,
                databricks_run_name=databricks_run_name,
            )
        except OperationFailed:
            databricks_run = workspace.jobs.get_run(run_id=waiter.run_id)
            task_run_id = get_task(databricks_run=databricks_run).run_id
            error_dict = workspace.jobs.get_run_output(task_run_id).as_dict()
            error_trace = error_dict.pop("error_trace", "")
            custom_error = "Error information and metadata:\n"
            custom_error += json.dumps(error_dict, indent=1)
            custom_error += "\nError trace from databricks:\n" if error_trace else ""
            custom_error += error_trace
            raise MLRunRuntimeError(custom_error)
        finally:
            save_credentials(
                workspace=workspace,
                waiter=waiter,
                host=host,
                token=databricks_token,
                cluster_id=cluster_id,
                is_finished=True,
            )

        task_run_id = get_task(run).run_id
        run_output = workspace.jobs.get_run_output(task_run_id)
    finally:
        workspace.dbfs.delete(script_path_on_dbfs)
        workspace.dbfs.delete(artifact_json_path)

    #  This code will not run in the case of an exception, within the outer try-finally block:
    logger.info(f"Job finished: {run.run_page_url}")
    logger.info(f"Logs:\n{run_output.logs}")
    run_output_dict = run_output.as_dict()
    run_output_dict.pop("logs", None)
    context.log_artifact(
        "databricks_run_metadata",
        body=json.dumps(run_output_dict),
        format="json",
    )

    run_result_state = run_output.metadata.state.result_state
    if run_result_state == RunResultState.CANCELED:
        raise MLRunTaskCancelledError(f"Task {task_run_id} has been cancelled")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from nuclio.build import mlrun_footer

import mlrun
import mlrun.utils.helpers

from ..model import ModelObj
from ..utils import generate_object_uri
from .utils import enrich_function_from_dict


class FunctionReference(ModelObj):
    """function reference/template, point to function and add/override resources"""

    def __init__(
        self,
        url=None,
        image=None,
        requirements=None,
        code=None,
        spec=None,
        kind=None,
        name=None,
        track_models=None,
    ):
        self.url = url
        self.kind = kind
        self.image = image
        self.requirements = requirements
        self.name = name
        if hasattr(spec, "to_dict"):
            spec = spec.to_dict()
        self.spec = spec
        self.code = code
        self.track_models = track_models

        self._function = None
        self._address = None

    def is_empty(self):
        if self.url or self.code or self.spec:
            return False
        return True

    def fullname(self, parent):
        return f"{parent.metadata.name}-{self.name}"

    def uri(self, parent, tag=None, hash_key=None, fullname=True):
        name = self.fullname(parent) if fullname else self.name
        return generate_object_uri(
            parent.metadata.project,
            name,
            tag=tag or parent.metadata.tag,
            hash_key=hash_key,
        )

    @property
    def function_object(self):
        """get the generated function object"""
        return self._function

    def to_function(self, default_kind=None, default_image=None):
        """generate a function object from the ref definitions"""
        if self.url and "://" not in self.url:
            if not os.path.isfile(self.url):
                raise OSError(f"{self.url} not found")

        kind = self.kind or default_kind
        if self.url:
            if (
                self.url.endswith(".yaml")
                or self.url.startswith("db://")
                or self.url.startswith("hub://")
            ):
                func = mlrun.import_function(self.url)
                func.spec.image = self.image or func.spec.image or default_image
            elif self.url.endswith(".ipynb"):
                func = mlrun.code_to_function(
                    self.name, filename=self.url, image=self.image, kind=kind
                )
                func.spec.image = func.spec.image or default_image
            elif self.url.endswith(".py"):
                # todo: support code text as input (for UI)
                image = self.image or default_image
                if not image:
                    raise ValueError(
                        "image must be provided with py code files, "
                        "use function object for more control/settings"
                    )
                func = mlrun.code_to_function(
                    self.name, filename=self.url, image=image, kind=kind
                )
            else:
                raise ValueError(f"unsupported function url {self.url} or no spec")
            if self.spec:
                func = enrich_function_from_dict(func, self.spec)
        elif self.code is not None:
            code = self.code
            if kind == mlrun.runtimes.RuntimeKinds.serving:
                code = code + mlrun_footer.format(
                    mlrun.runtimes.nuclio.serving.serving_subkind
                )
            func = mlrun.new_function(
                self.name, kind=kind, image=self.image or default_image
            )
            data = mlrun.utils.helpers.encode_user_code(code)
            func.spec.build.functionSourceCode = data
            if kind not in mlrun.runtimes.RuntimeKinds.nuclio_runtimes():
                func.spec.default_handler = "handler"
            if self.spec:
                func = enrich_function_from_dict(func, self.spec)
        elif self.spec:
            func = mlrun.new_function(self.name, runtime=self.spec)
        else:
            raise ValueError("url or spec or code must be specified")

        if self.requirements:
            func.with_requirements(self.requirements)
        self._function = func
        func.spec.track_models = self.track_models
        return func

    @property
    def address(self):
        return self._address

    def deploy(self, **kwargs):
        """deploy the function"""
        self._address = self._function.deploy(**kwargs)
        return self._address



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import plotly.graph_objects as go
from sklearn.calibration import calibration_curve

from mlrun.artifacts import Artifact, PlotlyArtifact

from ..plan import MLPlanStages, MLPlotPlan
from ..utils import MLTypes


class CalibrationCurvePlan(MLPlotPlan):
    """
    Plan for producing a calibration curve - computed true and predicted probabilities for a calibration curve. The
    method assumes the inputs come from a binary classifier, and discretize the [0, 1] interval into bins.
    """

    _ARTIFACT_NAME = "calibration-curve"

    def __init__(
        self,
        n_bins: int = 5,
        strategy: str = "uniform",
    ):
        """
        Initialize a calibration curve plan with the given configuration.

        To read more about the parameters, head to the SciKit-Learn docs at:
        https://scikit-learn.org/stable/modules/generated/sklearn.calibration.calibration_curve.html

        :param n_bins:    Number of bins to discretize the [0, 1] interval.
        :param strategy:  Strategy used to define the widths of the bins. Can be on of {‘uniform’, ‘quantile’}.
                          Default: "uniform".
        """
        # Store the parameters:
        self._n_bins = n_bins
        self._strategy = strategy

        # Continue the initialization for the MLPlan:
        super().__init__(need_probabilities=True)

    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether or not the plan is fit for production by the given stage and prediction probabilities. The
        calibration curve is ready only post prediction.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        return stage == MLPlanStages.POST_PREDICT and is_probabilities

    def produce(
        self,
        y: MLTypes.DatasetType,
        y_pred: MLTypes.DatasetType = None,
        model: MLTypes.ModelType = None,
        x: MLTypes.DatasetType = None,
        **kwargs,
    ) -> dict[str, Artifact]:
        """
        Produce the calibration curve according to the ground truth (y) and predictions (y_pred) values. If predictions
        are not available, the model and a dataset can be given to produce them.

        :param y:      The ground truth values.
        :param y_pred: The predictions values.
        :param model:  Model to produce the predictions.
        :param x:      Input dataset to produce the predictions.

        :return: The produced calibration curve artifact in an artifacts dictionary.
        """
        # Calculate the calibration curve:
        prob_true, prob_pred = calibration_curve(
            y,
            y_pred[:, -1],  # Take only the second class probabilities (1, not 0).
            n_bins=self._n_bins,
            strategy=self._strategy,
        )

        # Create the figure:
        fig = go.Figure(
            data=[go.Scatter(x=prob_true, y=prob_pred)],
            layout={"title": {"text": "Calibration Curve"}},
        )

        # Add custom x-axis title:
        fig.add_annotation(
            {
                "font": {"color": "black", "size": 14},
                "x": 0.5,
                "y": -0.15,
                "showarrow": False,
                "text": "prob_true",
                "xref": "paper",
                "yref": "paper",
            }
        )

        # Add custom y-axis title:
        fig.add_annotation(
            {
                "font": {"color": "black", "size": 14},
                "x": -0.1,
                "y": 0.5,
                "showarrow": False,
                "text": "prob_pred",
                "textangle": -90,
                "xref": "paper",
                "yref": "paper",
            }
        )

        # Adjust margins to make room for yaxis title
        fig.update_layout(margin={"t": 100, "l": 100}, width=800, height=500)

        # Creating the artifact:
        self._artifacts[self._ARTIFACT_NAME] = PlotlyArtifact(
            key=self._ARTIFACT_NAME,
            figure=fig,
        )

        return self._artifacts



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from urllib.parse import urlparse

from fsspec import AbstractFileSystem


class WasbFS(AbstractFileSystem):
    protocol = "wasb"

    def __init__(
        self,
        account_name: Optional[str] = None,
        account_key: Optional[str] = None,
        connection_string: Optional[str] = None,
        credential: Optional[str] = None,
        sas_token: Optional[str] = None,
        request_session=None,
        socket_timeout: Optional[int] = None,
        blocksize: Optional[int] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        tenant_id: Optional[str] = None,
        anon: bool = True,
        location_mode: Optional[str] = None,
        loop=None,
        asynchronous: bool = False,
        default_fill_cache: bool = True,
        default_cache_type: Optional[str] = None,
        **kwargs,
    ):
        from adlfs import AzureBlobFileSystem

        self.azure_blob_fs = AzureBlobFileSystem(
            account_name,
            account_key,
            connection_string,
            credential,
            sas_token,
            request_session,
            socket_timeout,
            blocksize,
            client_id,
            client_secret,
            tenant_id,
            anon,
            location_mode,
            loop,
            asynchronous,
            default_fill_cache,
            default_cache_type,
            **kwargs,
        )

    @staticmethod
    def _convert_wasb_schema_to_az(url):
        # convert wasbs schema url to az schema url. Used before passing the url to AzureBlobFS.
        # wasbs://mycontainer@myaccount/path/to/obj is equivalent to
        # az://mycontainer/path/to/obj
        az_path = url
        parsed_url = urlparse(url)
        if parsed_url.scheme:
            if (
                parsed_url.scheme.lower() == "wasb"
                or parsed_url.scheme.lower() == "wasbs"
            ):
                az_path = (
                    "az://"
                    + parsed_url.username
                    + ("/" if not parsed_url.path.startswith("/") else parsed_url.path)
                )
            else:
                raise ValueError("Operation expects to wasb or wasbs scheme only!")
        return az_path

    def info(self, path, refresh=False, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.info(az_path, refresh, **kwargs)

    def glob(self, path, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.glob(az_path, **kwargs)

    def ls(
        self,
        path: str,
        detail: bool = False,
        invalidate_cache: bool = False,
        delimiter: str = "/",
        return_glob: bool = False,
        **kwargs,
    ):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.ls(
            az_path,
            detail,
            invalidate_cache,
            delimiter,
            return_glob,
            **kwargs,
        )

    def find(self, path, withdirs=False, prefix="", **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.find(az_path, withdirs, prefix, **kwargs)

    def mkdir(self, path, exist_ok=False):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.mkdir(az_path, exist_ok)

    def rmdir(self, path: str, delimiter="/", **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.rmdir(az_path, delimiter, **kwargs)

    def size(self, path):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.size(az_path)

    def isfile(self, path):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.isfile(az_path)

    def isdir(self, path):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.isdir(az_path)

    def exists(self, path):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.exists(az_path)

    def cat(self, path, recursive=False, on_error="raise", **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.cat(az_path, recursive, on_error, **kwargs)

    def url(self, path, expires=3600, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.url(az_path, expires, **kwargs)

    def expand_path(self, path, recursive=False, maxdepth=None):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.expand_path(az_path, recursive, maxdepth)

    def rm(self, path, recursive=False, maxdepth=None, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.rm(az_path, recursive, maxdepth, **kwargs)

    def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.open(
            az_path, mode, block_size, cache_options, **kwargs
        )

    def touch(self, path, truncate=True, **kwargs):
        az_path = self._convert_wasb_schema_to_az(path)
        return self.azure_blob_fs.touch(az_path, truncate, **kwargs)

    @classmethod
    def _strip_protocol(cls, path: str):
        az_path = cls._convert_wasb_schema_to_az(path)
        from adlfs import AzureBlobFileSystem

        return AzureBlobFileSystem._strip_protocol(az_path)

    @staticmethod
    def _get_kwargs_from_urls(paths):
        from adlfs import AzureBlobFileSystem

        return AzureBlobFileSystem._get_kwargs_from_urls(paths)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime
from typing import Optional, Union

import numpy as np
import pandas as pd
from IPython.display import HTML, display
from pandas.api.types import is_numeric_dtype, is_string_dtype

import mlrun
from mlrun.utils import filter_warnings, flatten

max_table_rows = 50


def _gen_dropdown_buttons(output_cols) -> list:
    """Uses each output col name to generate an equivalent dropdown button for plotting."""

    output_cols = [col[len("output.") :] for col in output_cols]

    def gen_bool_list(col):
        return [name == col for name in output_cols]

    buttons = [
        dict(
            label=col,
            method="update",
            args=[{"visible": gen_bool_list(col)}],
        )
        for col in output_cols
    ]

    return buttons


def _gen_dimensions(
    df: pd.DataFrame, col: str, prefix: Optional[str] = None, is_index=False
) -> dict:
    """
    Computes the plotting dimensions of each parameter/output col according to its type.
    :param df: Dataframe containing the data to be plotted
    :param col: Column name for which its dimensions will be computed
    :returns dimension: Dimensions and instructions required to plot a parameter.
    """

    dimension = {}

    if is_numeric_dtype(df[col]):
        dimension["range"] = [min(df[col]), max(df[col])]
        dimension["values"] = df[col]

    elif is_string_dtype(df[col]):
        dimension["range"] = [0, len(df[col])]
        dimension["values"] = [(list(df[col]).index(val)) for val in list(df[col])]
        dimension["tickvals"] = [(list(df[col]).index(val)) for val in list(df[col])]
        dimension["ticktext"] = list(df[col])

    if col == "iter" or is_index:
        dimension["tickvals"] = (np.arange(1, max(df[col] + 1))).tolist()

    # Axis name
    if prefix:
        col = col[len(prefix) :]
    dimension["label"] = col

    return dimension


def _drop_identical(df: pd.DataFrame, param_cols) -> pd.DataFrame:
    """Drop columns with identical values throughout iterations (no variance)."""

    dropped_cols = []
    for col in param_cols:
        if df[col].nunique() == 1:
            df = df.drop(columns=[col])
            dropped_cols.append(col)
    return df, [col for col in param_cols if col not in dropped_cols]


def _get_column_names(df: pd.DataFrame):
    params = []
    outputs = []
    for name in df.columns:
        if name.startswith("param."):
            params.append(name)
        if name.startswith("output."):
            outputs.append(name)
    return params, outputs


@filter_warnings("ignore", FutureWarning)
def gen_pcp_plot(
    source_df: pd.DataFrame,
    index_col: str,
    hide_identical: bool = True,
    exclude: Optional[list] = None,
    colorscale: Optional[str] = None,
):
    """
    Creates a list composed of the data to be plotted as a Parallel Coordinate, this includes
    the dimensions, dropdown buttons, and visibility of plots.
    :param source_df: Result of the hyperparameter run as a Dataframe
    :param index_col: index column name
    :param hide_identical: Ignores parameters that remain the same throughout iterations
    :param exclude: User-provided list of parameters to be excluded from the graph
    :param colorscale: colors used for the lines in the parallel coordinate plot
    :returns plot_as_html: The Parallel Coordinate plot in HTML format.
    """

    try:
        import plotly.graph_objects as go
    except ImportError as exc:
        raise ImportError(f"{exc}, please run 'pip install plotly'")

    data = []

    # Remove unwanted columns
    for col in exclude or []:
        if f"param.{col}" in source_df.columns:
            source_df = source_df.drop(columns=[f"param.{col}"])

    param_cols, output_cols = _get_column_names(source_df)

    # Drop identical columns
    if hide_identical:
        source_df, param_cols = _drop_identical(source_df, param_cols)

    for output in output_cols:
        # Creating Axes and Dimensions
        dimensions = [(_gen_dimensions(source_df, col, "param.")) for col in param_cols]
        dimensions.insert(0, _gen_dimensions(source_df, index_col))
        dimensions.append(_gen_dimensions(source_df, output, "output."))

        # Plot Visibility on Dropdown
        visibility = True if output == output_cols[0] else False

        # Appending to list
        data.append(
            go.Parcoords(
                line=dict(
                    color=source_df[index_col], colorscale=colorscale or "viridis"
                ),
                dimensions=dimensions,
                visible=visibility,
            )
        )

    # Set position of dropdown buttons
    layout = dict(
        showlegend=True,
        updatemenus=[
            dict(
                active=0,
                xanchor="left",
                x=0.15,
                yanchor="bottom",
                y=1.3,
                bgcolor="#ffffff",
                showactive=True,
                buttons=_gen_dropdown_buttons(output_cols),
            )
        ],
    )

    fig = go.Figure(data=data, layout=layout)

    # Workaround to remove dropdown button opacity
    fig.update_layout(autosize=True, margin=dict(l=50, r=50, b=50, t=50, pad=4))

    # Add annotation to dropdown
    fig.update_layout(
        annotations=[
            dict(
                text="Output metric:",
                x=0.0,
                xref="paper",
                y=1.4,
                yref="paper",
                align="right",
                showarrow=False,
            ),
        ]
    )

    return fig.to_html()


def _show_and_export_html(html: str, show=None, filename=None, runs_list=None):
    if runs_list:
        html_table = runs_list.show(False, short=True)

    if filename:
        dir = os.path.dirname(filename)
        if dir:
            os.makedirs(dir, exist_ok=True)
        with open(filename, "w") as fp:
            if runs_list:
                # add runs table after the plot
                fp.write(html[: html.rfind("</body>")])
                fp.write("<br>" + html_table)
                fp.write("</body></html>")
            else:
                fp.write(html)
    if show or (show is None and mlrun.utils.is_jupyter):
        display(HTML(html))
        if runs_list and len(runs_list) <= max_table_rows:
            display(HTML(html_table))
    else:
        return html


def _runs_list_to_df(runs_list, extend_iterations=False):
    runs_df = runs_list.to_df(extend_iterations=extend_iterations)

    # Remove empty param/result runs
    runs_df = runs_df[(runs_df["parameters"].str.len() >= 1)]
    runs_df = runs_df[(runs_df["results"].str.len() >= 1)]

    runs_df = flatten(runs_df, "parameters", "param.")
    runs_df = flatten(runs_df, "results", "output.")
    runs_df["iter"] = range(1, 1 + len(runs_df))
    return runs_df


@filter_warnings("ignore", FutureWarning)
def compare_run_objects(
    runs_list: Union[mlrun.model.RunObject, list[mlrun.model.RunObject]],
    hide_identical: bool = True,
    exclude: Optional[list] = None,
    show: Optional[bool] = None,
    extend_iterations=True,
    filename=None,
    colorscale: Optional[str] = None,
):
    """return/show parallel coordinates plot + table to compare between a list of runs or run iterations

    example:

        # create two runs and compare them
        fn = mlrun.code_to_function("myfunc", filename="./code.py", image='mlrun/mlrun', handler="handler", kind='job')
        r1 = fn.run(name="r1", params={"p1": 2, "p2":3})
        r2 = fn.run(name="r2", params={"p1": 3, "p2":5})
        compare_run_objects([r1, r2], filename="comp.html")

        # compare hyper-parameter run
        run_spec = mlrun.new_task("r1").with_hyper_params({"p1": [1, 5, 3]}, selector="max.accuracy")
        run = fn.run(run_spec)
        compare_run_objects(run, hide_identical=False)

    :param runs_list:      List or instance of MLRun RunObject (result of function.run())
    :param hide_identical: hide columns with identical values
    :param exclude:        User-provided list of parameters to be excluded from the plot
    :param show:           Allows the user to display the plot within the notebook
    :param extend_iterations: include the iteration (hyper-param) results
    :param filename:       Output filename to save the plot html file
    :param colorscale:     colors used for the lines in the parallel coordinate plot
    :return:  plot html
    """

    if isinstance(runs_list, mlrun.model.RunObject):
        runs_list = [runs_list]
    if isinstance(runs_list, list):
        runs_list = mlrun.lists.RunList([run.to_dict() for run in runs_list])
    source_df = _runs_list_to_df(runs_list, extend_iterations)
    plot_as_html = gen_pcp_plot(
        source_df,
        index_col="iter",
        hide_identical=hide_identical,
        exclude=exclude,
        colorscale=colorscale,
    )
    return _show_and_export_html(plot_as_html, show, filename, runs_list=runs_list)


@filter_warnings("ignore", FutureWarning)
def compare_db_runs(
    project_name=None,
    run_name=None,
    labels=None,
    iter=False,
    start_time_from: Optional[datetime] = None,
    hide_identical: bool = True,
    exclude: Optional[list] = None,
    show=None,
    colorscale: str = "Blues",
    filename=None,
    **query_args,
) -> str:
    """
    Get the selected list of runs from MLRun DB and return/show a parallel coordinates plots + table.

    example:

        compare_db_runs("my-project", run_name="train")

    :param project_name: Project that the runs belongs to
    :param run_name: Name of the run to retrieve
    :param labels: List runs that have a specific label assigned. Currently only a single label filter can be
            applied, otherwise result will be empty.
    :param iter:            If ``True`` return runs from all iterations. Otherwise, return only parent runs (iter 0).
    :param start_time_from: Filter by run start time in ``[start_time_from, start_time_to]``.
    :param hide_identical:  hide columns with identical values
    :param exclude:         User-provided list of parameters to be excluded from the plot
    :param show:            Allows the user to display the plot within the notebook
    :param filename:        Output filename to save the plot html file
    :param colorscale:      colors used for the lines in the parallel coordinate plot
    :param query_args:      additional list_runs() query arguments
    :return:  plot html
    """

    runs_list = mlrun.get_run_db().list_runs(
        labels=labels,
        iter=iter,
        start_time_from=start_time_from,
        name=run_name,
        project=project_name,
        **query_args,
    )

    exclude = exclude or []
    runs_df = _runs_list_to_df(runs_list)
    plot_as_html = gen_pcp_plot(
        runs_df,
        index_col="iter",
        hide_identical=hide_identical,
        exclude=exclude,
        colorscale=colorscale,
    )
    return _show_and_export_html(plot_as_html, show, filename, runs_list=runs_list)
